from typing import Dict,Any,TYPE_CHECKING
import numpy as np
from sklearn import metrics

from nlp_tools.types import MultiLabelClassificationLabelVar

if TYPE_CHECKING:
    from nlp_tools.utils import MultiLabelBinarizer

def multi_label_classification_report(y_true: MultiLabelClassificationLabelVar,
                                      y_pred: MultiLabelClassificationLabelVar,
                                      *,
                                      binarizer: 'MultiLabelBinarizer',
                                      digits: int = 4,
                                      verbose: int = 1) -> Dict[str, Any]:
    y_pred_b = binarizer.transform(y_pred)
    y_true_b = binarizer.transform(y_true)

    report_dic: Dict = {}
    details: Dict = {}

    rows = []
    ps, rs, f1s, s = [], [], [], []
    for c_index, c in enumerate(binarizer.classes):
        precision = metrics.precision_score(y_true_b[:, c_index], y_pred_b[:, c_index])
        recall = metrics.recall_score(y_true_b[:, c_index], y_pred_b[:, c_index])
        f1 = metrics.f1_score(y_true_b[:, c_index], y_pred_b[:, c_index])
        support = len(np.where(y_true_b[:, c_index] == 1)[0])
        details[c] = {
            'precision': precision,
            'recall': recall,
            'f1': f1,
            'support': support
        }

        rows.append((c, precision, recall, f1, support))
        ps.append(precision)
        rs.append(recall)
        f1s.append(f1)
        s.append(support)

    report_dic['precision'] = np.average(ps, weights=s)
    report_dic['recall'] = np.average(rs, weights=s)
    report_dic['f1-score'] = np.average(f1s, weights=s)
    report_dic['support'] = np.sum(s)

    headers = ["precision", "recall", "f1-score", "support"]
    head_fmt = '{:>{width}s} ' + ' {:>9}' * len(headers) + '\n'

    report = head_fmt.format('', *headers, width=20)

    row_fmt = '{:>{width}s}  {:>9.{digits}f} {:>9.{digits}f} {:>9.{digits}f} {:>9}\n'

    for row in rows:
        report += row_fmt.format(*row, width=20, digits=digits)

    # compute averages
    report += row_fmt.format('macro avg',
                             np.average(ps, weights=s),
                             np.average(rs, weights=s),
                             np.average(f1s, weights=s),
                             np.sum(s),
                             width=20, digits=digits)

    report_dic['detail'] = details
    print(report)

    return report_dic


if __name__ == "__main__":
    pass



from abc import ABC
from typing import Dict,List,Optional,Any,Tuple
import numpy as np

from nlp_tools.generators import  CorpusGenerator
from nlp_tools.types import TextSamplesVar
from nlp_tools.tokenizer import ABCTokenizer
from nlp_tools.utils import load_data_object


class ABCSentenceProcessor(ABC):
    def to_dict(self) -> Dict[str, Any]:
        return {
            'config': {
                'max_sentence_length':self.max_sentence_length,
                "embedding_max_position":self.embedding_max_position,
            },
            'text_tokenizer': self.text_tokenizer.to_dict(),

            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
        }

    def __init__(self,text_tokenizer=None, max_sentence_length=None,embedding_max_position=None,**kwargs: Any) -> None:
        self.text_tokenizer: ABCTokenizer = text_tokenizer
        self.max_sentence_length = max_sentence_length
        self.embedding_max_position = embedding_max_position



    def _override_load_model(self,config:Dict) -> None:
        self.text_tokenizer: ABCTokenizer = load_data_object(config['text_tokenizer'])


    def build_vocab(self,
                    x_data: TextSamplesVar,
                    y_data: TextSamplesVar) -> None:
        corpus_gen = CorpusGenerator(x_data, y_data)
        self.build_vocab_generator([corpus_gen])

    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        raise NotImplementedError

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        return 2, batch_size, seq_length


    def transform(self,samples,seq_length: int = None):
        raise NotImplementedError

    def inverse_transform(self,
                          labels: List[int],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[str]:
        raise NotImplementedError






import re
from nlp_tools import utils
from nlp_tools.utils.text_clean import _regular_
from typing import Tuple, List
import logging
def _good_line_(line):
    """
    判断一句话是否是好的语料,即判断
    :param line:
    :return:
    """
    if len(line) == 0:
        return False
    ch_count = 0
    for c in line:
        # 中文字符范围
        if '\u4e00' <= c <= '\u9fff':
            ch_count += 1
    if ch_count / float(len(line)) >= 0.8 and len(re.findall(r'[a-zA-Z0-9]', ''.join(line))) < 3 \
            and len(re.findall(r'[ˇˊˋˍεπのゞェーω]', ''.join(line))) < 3:
        return True
    return False


class XiaoHuangJi(object):
    __corpus_name__ = 'xiao_huang_ji'

    @classmethod
    def load_data(cls,
                  file_path: str ,
                  shuffle: bool = True) -> Tuple[List[List[str]], List[List[str]]]:
        """
                Load dataset as sequence labeling format, char level tokenized

                features: ``[['海', '钓', '比', '赛', '地', '点', '在', '厦', '门', ...], ...]``

                labels: ``[[海', '钓', '比', '赛', '地', '点', , ...], ...]``

                Args:
                    file_path: file ablsute path
                    shuffle: should shuffle or not, default True.

                Returns:
                    dataset_features and dataset labels
                """
        x_data = []
        y_data = []
        q = None
        with open(file_path, 'r', encoding='utf-8') as fr:
            for line in fr:
                if line.startswith('M '):
                    if q is None:
                        q = _regular_(line[2:-1])
                    else:
                        a = _regular_(line[2:-1])
                        if _good_line_(q) and _good_line_(a):
                            x_data.append(list(q))
                            y_data.append(list(a))
                        q = None

        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)
        logging.debug(f"loaded {len(x_data)} samples from {file_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")
        return x_data, y_data

if __name__ == '__main__':
    file_path = r'F:\nlp-data\chat\xiaohuangji50w_nofenci.conv'
    x,y = XiaoHuangJi.load_data(file_path)
    print(x[0])
    print(y[0])


# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: abc_model.py
# time: 4:30 下午

from abc import ABC
from typing import List, Dict, Any, Union, Optional

import numpy as np
import tensorflow as tf

import nlp_tools
from nlp_tools.embeddings import ABCEmbedding, BareEmbedding,TransformerEmbedding,BertEmbedding
from nlp_tools.generators import CorpusGenerator, BatchGenerator
from nlp_tools.logger import logger
from nlp_tools.metrics.sequence_labeling import get_entities
from nlp_tools.metrics.sequence_labeling import sequence_labeling_report
from nlp_tools.processors import SequenceProcessor
from nlp_tools.tasks.abs_task_model import ABCTaskModel
from nlp_tools.types import TextSamplesVar
from nlp_tools.optimizer import MultiOptimizer

from tensorflow.keras.optimizers import Adam
from nlp_tools.utils.ner_utils import output_ner_results
from nlp_tools.utils.training_tips import creat_FGM
from nlp_tools.loss.r_drop_loss import RDropLoss
from sklearn.model_selection._split import StratifiedKFold,KFold
from tensorflow.keras import backend as K
import functools
import os



class ABCLabelingModel(ABCTaskModel, ABC):
    """
    Abstract Labeling Model
    """

    def __init__(self,
                 text_processor,
                 label_processor,
                 embedding: ABCEmbedding = None,
                 max_sequence_length: int = None,
                 hyper_parameters: Dict[str, Dict[str, Any]] = None,
                 train_sequece_length_as_max_sequence_length = False,
                 use_FGM = True,
                 use_rdrop=True):
        """

        Args:
            embedding: embedding object
            max_sequence_length: target sequence length
            hyper_parameters: hyper_parameters to overwrite
        """
        super(ABCLabelingModel, self).__init__()
        if hyper_parameters is None:
            hyper_parameters = self.default_hyper_parameters()

        self.tf_model: Optional[tf.keras.Model] = None
        self.embedding = embedding
        self.hyper_parameters = hyper_parameters


        self.text_processor = text_processor
        self.label_processor = label_processor

        self.max_sequence_length = max_sequence_length
        self.train_max_sequence_length = None
        self.train_sequece_length_as_max_sequence_length = train_sequece_length_as_max_sequence_length
        self.use_FGM = use_FGM
        self.use_rdrop = use_rdrop

        ## 如果分词器支持keep_tokens，且embedding也支持keep_tokens，则需要吧分词器的keep_tokens跟新到
        if self.text_processor:
            if hasattr(self.text_processor.text_tokenizer,"keep_tokens") and self.text_processor.text_tokenizer.keep_tokens != "" and hasattr(self.embedding,"keep_tokens"):
                self.embedding.keep_tokens = self.text_processor.text_tokenizer.keep_tokens


        self.embedding_max_position = None
        if self.embedding and  hasattr(self.embedding,"max_position"):
            self.embedding_max_position = self.embedding.max_position



    def build_model(self,
                    x_data: TextSamplesVar,
                    y_data: TextSamplesVar) -> None:
        """
        Build Model with x_data and y_data

        This function will setup a :class:`CorpusGenerator`,
         then call :meth:`ABCClassificationModel.build_model_gen` for preparing processor and model

        Args:
            x_data:
            y_data:

        Returns:

        """

        train_gen = CorpusGenerator(x_data, y_data)
        self.build_model_generator([train_gen])

    def build_model_generator(self,
                              generators: List[CorpusGenerator]) -> None:

        self.label_processor.build_vocab_generator(generators)
        if self.train_max_sequence_length is None and self.text_processor:
            self.train_max_sequence_length = self.text_processor.get_max_seq_length_from_corpus(generators,cover_rate=1.0)

        if self.train_sequece_length_as_max_sequence_length:
            self.max_sequence_length = self.train_sequece_length_as_max_sequence_length

        self.label_processor.update_length_info(embedding_max_position=self.embedding_max_position,max_sentence_length=self.max_sequence_length)

        if self.text_processor:
            self.text_processor.update_length_info(embedding_max_position=self.embedding_max_position,max_sentence_length=self.max_sequence_length)

        if self.embedding:
            self.embedding.build_embedding_model()


        if self.tf_model is None:
            self.build_model_arc()
            self.compile_model()

            # 替换model.train_step 方法即可,并且删除原有的 train_function方法
            if self.use_FGM:
                train_step = creat_FGM()
                self.tf_model.train_step = functools.partial(train_step, self.tf_model)
                self.tf_model.train_function = None


    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            loss = 'sparse_categorical_crossentropy'


        if type(self.embedding) in [TransformerEmbedding,BertEmbedding] :
            total_layers = self.tf_model.layers
            transfomer_layers = self.embedding.embed_model.layers
            no_transformer_layers = [layer for layer in total_layers if layer not in transfomer_layers]
            optimizer_list = [
                Adam(learning_rate=1e-5),
                Adam(learning_rate=2e-5)
            ]
            optimizers_and_layers = [(optimizer_list[0], no_transformer_layers), (optimizer_list[1], transfomer_layers)]
            optimizer = MultiOptimizer(optimizers_and_layers)




        if optimizer is None:
            optimizer = Adam(learning_rate=1e-5)
        if self.use_rdrop:
            if type(loss) == list:
                loss = [RDropLoss(i) for i in loss]
            else:
                loss = RDropLoss(loss)


        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)



    def fit(self,
            train_data,
            validate_data=None,
            batch_size: int = 64,
            epochs: int = 5,
            callbacks: List[tf.keras.callbacks.Callback] = None,
            fit_kwargs: Dict = None,
            generator=None) -> 'tf.keras.callbacks.History':

        # 修正一下数据集，因为经过分词以后，实体的位置的index可能发现改变，需要修正一下
        train_data = self.ner_data_tokeninzed_and_change_label(train_data)
        if validate_data:
            validate_data = self.ner_data_tokeninzed_and_change_label(validate_data)

        if fit_kwargs is None:
            fit_kwargs = {}

        if generator == None:
            generator = BatchGenerator

        train_generator = generator(train_data,
                                   text_processor=self.text_processor,
                                   label_processor=self.label_processor,
                                   seq_length=self.train_max_sequence_length,
                                   batch_size=batch_size,
                                    use_rdrop=self.use_rdrop)
        if validate_data:
            valid_generator = generator(validate_data,
                                     text_processor=self.text_processor,
                                     label_processor=self.label_processor,
                                     seq_length=self.train_max_sequence_length,
                                     batch_size=batch_size,
                                     use_rdrop=self.use_rdrop)
            fit_kwargs['validation_data'] = valid_generator.forfit()
            fit_kwargs['validation_steps'] = len(valid_generator)
        else:
            valid_generator = None


        return self.fit_generator(train_sample_gen=train_generator,
                                  valid_sample_gen=valid_generator,
                                  epochs=epochs,
                                  callbacks=callbacks,
                                  fit_kwargs=fit_kwargs)

    def fit_generator(self,
                      train_sample_gen,
                      valid_sample_gen,
                      epochs: int = 5,
                      callbacks: List['tf.keras.callbacks.Callback'] = None,
                      fit_kwargs: Dict = None) -> 'tf.keras.callbacks.History':
        self.build_model_generator([g for g in [train_sample_gen, valid_sample_gen] if g])
        return self.tf_model.fit(train_sample_gen.forfit(),
                                 steps_per_epoch=len(train_sample_gen),
                                 epochs=epochs,
                                 callbacks=callbacks,
                                 **fit_kwargs)

    def predict(self,
                x_data: TextSamplesVar,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs: Dict = None) -> List[List[str]]:
        """
        Generates output predictions for the input samples.

        Computation is done in batches.

        Args:
            x_data: The input data, as a Numpy array (or list of Numpy arrays if the model has multiple inputs).
            batch_size: Integer. If unspecified, it will default to 32.
            truncating: remove values from sequences larger than `model.embedding.sequence_length`
            predict_kwargs: arguments passed to :meth:`tf.keras.Model.predict`

        Returns:
            array(s) of predictions.
        """
        if predict_kwargs is None:
            predict_kwargs = {}

        if truncating:
            seq_length = self.max_sequence_length
        else:
            seq_length = None

        if type(x_data[0]) == list:
            x_data = ["".join(x).replace("##","").replace('[CLS]',"").replace("[SEP]","") for x in x_data]
        x_data_tokenized = [self.text_processor.text_tokenizer.tokenize(x) for x in x_data]
        lengths = [len(x) for x in x_data_tokenized]

        tensor = self.text_processor.transform(x_data,seq_length=seq_length)
        logger.debug('predict seq_length: {}, input: {}'.format(seq_length, np.array(tensor).shape))
        pred = self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)
        pred = pred.argmax(-1)

        x_data_mapping = [self.text_processor.text_tokenizer.rematch(x, x_tokens) for x, x_tokens in
                          zip(x_data, x_data_tokenized)]

        res: List[List[str]] = self.label_processor.inverse_transform(pred,lengths=lengths,mapping_list=x_data_mapping)
        logger.debug('predict output: {}'.format(np.array(pred).shape))
        logger.debug('predict output argmax: {}'.format(pred))
        return res

    def predict_entities(self,
                         x_data,
                         batch_size: int = 32,
                         truncating: bool = False,
                         predict_kwargs: Dict = None):
        predict_result = self.predict(x_data, batch_size, truncating, predict_kwargs)

        result = output_ner_results(x_data, predict_result)
        return result

    def evaluate(self,data,
                 y_data,
                 batch_size: int = 32,
                 digits: int = 4,
                 truncating: bool = False) -> Dict:
        y_pred = self.predict(data,
                              batch_size=batch_size,
                              truncating=truncating)
        y_true = [seq[:len(y_pred[index])] for index, seq in enumerate(y_data)]

        new_y_pred = []
        for x in y_pred:
            new_y_pred.append([str(i) for i in x])
        new_y_true = []
        for x in y_true:
            new_y_true.append([str(i) for i in x])

        report = sequence_labeling_report(y_true, y_pred, digits=digits)
        return report

    def ner_data_tokeninzed_and_change_label(self,data):
        '''对训练数据分词处理，并跟新label的index'''
        if self.text_processor and hasattr(self.text_processor.text_tokenizer,'rematch'):
            tokenizer_data = []
            for (x, y) in data:
                item = []
                tokens = self.text_processor.text_tokenizer.tokenize(x)
                item.append(tokens)

                mapping = self.text_processor.text_tokenizer.rematch(x, tokens)
                start_mapping = {j[0]: i for i, j in enumerate(mapping) if j}
                end_mapping = {j[-1]: i for i, j in enumerate(mapping) if j}

                entities = []
                for start, end ,label in y:
                    if start in start_mapping and end in end_mapping:
                        start = start_mapping[start]
                        end = end_mapping[end]
                        entities.append((start,end,label))
                item.append(entities)

                tokenizer_data.append(item)
        else:
            tokenizer_data = data
        return tokenizer_data

    def smooth_labels(self, labels, factor=0.1):
        # smooth the labels
        labels *= (1 - factor)
        labels += (factor / labels.shape[1])

        return labels



from typing import Dict,Any,Optional

import numpy as np
from gensim.models import KeyedVectors
from tensorflow import keras

from nlp_tools.embeddings.abc_embedding import  ABCEmbedding
from nlp_tools.logger import logger

L = keras.layers


class WordEmbedding(ABCEmbedding):
    def to_dict(self) -> Dict[str, Any]:
        info_dic = super(WordEmbedding, self).to_dict()
        info_dic['config']['w2v_path'] = self.w2v_path
        info_dic['config']['w2v_kwargs'] = self.w2v_kwargs
        return info_dic

    def __init__(self,
                 w2v_path: str,
                 *,
                 w2v_kwargs: Dict[str, Any] = None,
                 **kwargs: Any):
        """
        Args:
            w2v_path: Word2Vec file path.
            w2v_kwargs: params pass to the ``load_word2vec_format()`` function
              of `gensim.models.KeyedVectors <https://radimrehurek.com/gensim/models/keyedvectors.html#module-gensim.models.keyedvectors>`_
            kwargs: additional params
        """
        if w2v_kwargs is None:
            w2v_kwargs = {}

        self.w2v_path = w2v_path
        self.w2v_kwargs = w2v_kwargs

        self.embedding_size = None
        self.w2v_matrix: np.ndarray = None



        super(WordEmbedding, self).__init__(**kwargs)

        self.vocab2idx = self.load_embed_vocab()

    def load_embed_vocab(self) -> Optional[Dict[str, int]]:
        w2v = KeyedVectors.load_word2vec_format(self.w2v_path, **self.w2v_kwargs)

        token2idx = {
            '[PAD]': 0,
            '[UNK]': 1,
            #'[CLS]': 2,
            #'[SEP]': 3
        }

        # if '[CLS]' not in w2v.vocab:
        #     token2idx['[CLS]'] = len(token2idx)
        #
        # if '[SEP]' not in w2v.vocab:
        #     token2idx['[SEP]'] = len(token2idx)

        for token in w2v.index2word:
            token2idx[token] = len(token2idx)

        vector_matrix = np.zeros((len(token2idx), w2v.vector_size))
        vector_matrix[1] = np.random.rand(w2v.vector_size)
        vector_matrix[2:] = w2v.vectors

        self.embedding_size = w2v.vector_size
        self.w2v_matrix = vector_matrix
        w2v_top_words = w2v.index2entity[:50]

        logger.debug('------------------------------------------------')
        logger.debug("Loaded gensim word2vec model's vocab")
        logger.debug('model        : {}'.format(self.w2v_path))
        logger.debug('word count   : {}'.format(len(self.w2v_matrix)))
        logger.debug('Top 50 words : {}'.format(w2v_top_words))
        logger.debug('------------------------------------------------')

        return token2idx

    def build_embedding_model(self,
                              *,
                              vocab_size: int = None,
                              force: bool = False,
                              **kwargs: Dict) -> None:
        if self.embed_model is None:
            input_tensor = L.Input(shape=(None,),
                                   name=f'input')
            layer_embedding = L.Embedding(len(self.vocab2idx),
                                          self.embedding_size,
                                          weights=[self.w2v_matrix],
                                          mask_zero=True,
                                          trainable=True,
                                          name=f'layer_embedding')

            embedded_tensor = layer_embedding(input_tensor)
            self.embed_model = keras.Model(input_tensor, embedded_tensor)


if __name__ == '__main__':
    word_vec_path = '/home/qiufengfeng/nlp/nlp_project/gitee_my/nlp_tools/example/pretraining/word2vec/model.vec'
    a = WordEmbedding(w2v_path=word_vec_path)





import tensorflow.keras.backend as K
import tensorflow as tf

class MaskLoss():
    def __init__(self,mask,loss_func):
        self.mask = mask
        self.loss_func = loss_func


    def __call__(self, y_true, y_pred):
        loss_ = self.loss_func(y_true,y_pred)
        mask = tf.cast(self.mask, dtype=loss_.dtype)  # 将前面统计的是否零转换成1，0的矩阵
        loss_ *= mask  # 将正常计算的loss加上mask的权重，就剔除了padding 0的影响
        loss_ = tf.math.divide_no_nan(tf.reduce_sum(loss_, axis=-1), tf.reduce_sum(mask, axis=-1))

        return  loss_


from typing import Dict,Any
from abc import ABC, abstractmethod
class ABCTokenizer(ABC):
    """Abstract base class for all implemented tokenizers.
    """

    def __init__(self):
        self._tokenizer = None

    @property
    def tokenizer(self):
        return self._tokenizer

    def to_dict(self) -> Dict[str, Any]:
        return {
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
            'config':{}
        }

    def tokenize(self, text,max_len=None,**kwargs):
        """
        Tokenize text into token sequence
        Args:
            text: target text sample

        Returns:
            List of tokens in this sample
        """
        raise NotImplementedError

    def encode(self,text,
                second_text=None,
                maxlen=None,**kwargs):
        raise NotImplementedError

    def id_to_token(self, id):
        raise  NotImplementedError




from tensorflow.keras.callbacks import Callback
import tensorflow.keras.backend as K
import numpy as np

class WarmupExponentialDecay(Callback):
    def __init__(self,lr_base=0.001,lr_min=0.0,decay=0.05,warmup_epochs=5):
        self.num_passed_batchs = 0   #一个计数器
        self.warmup_epochs=warmup_epochs
        self.lr=lr_base #learning_rate_base
        self.lr_min=lr_min #最小的起始学习率,此代码尚未实现
        self.decay=decay  #指数衰减率
        self.steps_per_epoch=0 #也是一个计数器
    def on_batch_begin(self, batch, logs=None):
        # params是模型自动传递给Callback的一些参数
        if self.steps_per_epoch==0:
            #防止跑验证集的时候呗更改了
            if self.params['steps'] == None:
                self.steps_per_epoch = np.ceil(1. * self.params['samples'] / self.params['batch_size'])
            else:
                self.steps_per_epoch = self.params['steps']
        if self.num_passed_batchs < self.steps_per_epoch * self.warmup_epochs:
            K.set_value(self.model.optimizer.lr,
                        0.1*(self.num_passed_batchs + 1) / self.steps_per_epoch / self.warmup_epochs)
        else:
            K.set_value(self.model.optimizer.lr,
                        self.lr*((1-self.decay)**(self.num_passed_batchs-self.steps_per_epoch*self.warmup_epochs)))
        self.num_passed_batchs += 1
    def on_epoch_begin(self,epoch,logs=None):
    #用来输出学习率的,可以删除
        print("learning_rate:",K.get_value(self.model.optimizer.lr))



import tensorflow as tf
from tensorflow.python.keras.engine import data_adapter
def creat_FGM(epsilon=1.0):
    @tf.function
    def train_step(self, data):
        '''
        计算在embedding上的gradient
        计算扰动 在embedding上加上扰动
        重新计算loss和gradient
        删除embedding上的扰动，并更新参数
        '''
        data = data_adapter.expand_1d(data)
        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)
        with tf.GradientTape() as tape:
            y_pred = self(x,training=True)
            loss = self.compiled_loss(
                y, y_pred, sample_weight, regularization_losses=self.losses)
        embedding = self.trainable_variables[0]
        embedding_gradients = tape.gradient(loss,[self.trainable_variables[0]])[0]
        embedding_gradients = tf.zeros_like(embedding) + embedding_gradients
        delta = epsilon * embedding_gradients / (tf.math.sqrt(tf.reduce_sum(embedding_gradients**2)) + 1e-8)  # 计算扰动
        self.trainable_variables[0].assign_add(delta)
        with tf.GradientTape() as tape2:
            y_pred = self(x,training=True)
            new_loss = self.compiled_loss(
                y, y_pred, sample_weight, regularization_losses=self.losses)
        gradients = tape2.gradient(new_loss,self.trainable_variables)
        self.trainable_variables[0].assign_sub(delta)

        self.optimizer.apply_gradients(zip(gradients,self.trainable_variables))
        #train_loss.update_state(loss)
        self.compiled_metrics.update_state(y, y_pred, sample_weight)
        return {m.name: m.result() for m in self.metrics}
    return train_step



# encoding: utf-8


import collections
import operator
from typing import Dict, List, Any, Optional, Union

import numpy as np
import tqdm
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.logger import logger
from nlp_tools.processors.abc_sentence_processor import ABCSentenceProcessor
from nlp_tools.types import TextSamplesVar


class SequenceProcessor(ABCSentenceProcessor):
    """
    Generic processors for the sequence samples.
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(SequenceProcessor, self).to_dict()
        data['config'].update({
        })
        return data

    def __init__(self,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        super(SequenceProcessor, self).__init__(**kwargs)



    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def transform(self,samples: TextSamplesVar,seq_length: int = None) :

        if seq_length is None and self.max_sentence_length is not None:
            seq_length = self.max_sentence_length

        if seq_length is None:
            seq_length = max([len(self.text_tokenizer.tokenize(i)) for i in samples])

        if self.embedding_max_position is not None and self.embedding_max_position < seq_length:
            seq_length = self.embedding_max_position



        tokenids, segment_ids = [],[]
        for seq in samples:
            seq_tokenids,seq_segment_ids = self.text_tokenizer.encode(seq,maxlen=seq_length)
            tokenids.append(seq_tokenids)
            segment_ids.append(seq_segment_ids)


        token_ids = pad_sequences(tokenids, seq_length, padding='post', truncating='post')
        segment_ids = pad_sequences(segment_ids, seq_length, padding='post', truncating='post')
        token_ids = np.array(token_ids)
        segment_ids = np.array(segment_ids)
        return token_ids, segment_ids


    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[List[str]]:
        result = []
        for index, seq in enumerate(labels):
            labels_ = []
            for idx in seq:
                labels_.append(self.text_tokenizer.id_to_token(idx))
            if lengths is not None:
                labels_ = labels_[1:lengths[index] + 1]
            else:
                labels_ = labels_[1:-1]
            result.append(labels_)
        return result

    def get_max_seq_length_from_corpus(self,
                                   generators: List[CorpusGenerator],
                                   cover_rate: float = 0.95) -> int:
        """
        Calculate proper sequence length according to the corpus

        Args:
            generators:
            use_label:
            cover_rate:

        Returns:

        """
        seq_lens = []
        for gen in generators:
            for (sentence, _) in tqdm.tqdm(gen.sample(), desc="Calculating sequence length"):
                seq_lens.append(len(self.text_tokenizer.tokenize(sentence)))
        if cover_rate == 1.0:
            target_index = -1
        else:
            target_index = int(cover_rate * len(seq_lens))
        sequence_length = sorted(seq_lens)[target_index]
        logger.debug(f'Calculated sequence length = {sequence_length}')
        return sequence_length



if __name__ == "__main__":
    pass



from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel
from nlp_tools.tasks.labeling.bi_lstm_model import BiLSTM_Model
from nlp_tools.tasks.labeling.bi_lstm_crf_model import BiLSTM_CRF_Model



from typing import Dict,Any
from tensorflow import keras
from tensorflow.keras import Sequential
from nlp_tools.layers import L
from nlp_tools.tasks.classification.abc_model import ABCClassificationModel
from tensorflow.keras.layers import *
from nlp_tools.layers.normal_attention import NormalAttentionLayer


class HierarchicalAttentionNetworks(ABCClassificationModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'sentence_attention':{
                'attention_dim': 128
            },
            'layer_bi_lstm': {
                'units': 10,
                'return_sequences': True
            },
            'layer_output': {

            },

            'dco_attention':{
                'attention_dim': 128
            },
        }

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model


        # build model structure in sequent way

        # sentEncoder = Sequential()
        # sentEncoder.add(embed_model)
        # sentEncoder.add(L.Bidirectional(L.GRU(**config['layer_bi_lstm'])))
        # sentEncoder.add(NormalAttentionLayer(name="sentence_attention",**config['sentence_attention']))
        # # sentence_bilstm = L.Bidirectional(L.LSTM(**config['layer_bi_lstm']))(embed_model.output)
        # # sentence_attention = NormalAttentionLayer(name="sentence_attention",**config['sentence_attention'])(sentence_bilstm)
        # # sentEncoder = keras.Model(self.embedding.embed_model.inputs, sentence_attention)
        #
        # review_input = L.Input(shape=(20, None), dtype='int32')
        #
        # review_encoder = L.TimeDistributed(sentEncoder)(review_input)
        #
        #
        #
        # doc_bilstm = L.Bidirectional(L.GRU(name='doc_lstm',**config['layer_bi_lstm']),name='doc_bi_lstm')(review_encoder)
        # doc_attention = NormalAttentionLayer(name="doc_attention",**config['dco_attention'])(doc_bilstm)
        # dense_output = L.Dense(output_dim, **config['layer_output'])(doc_attention)
        # final_output = self._activation_layer()(dense_output)

        from tensorflow.keras import Model
        MAX_SENT_LENGTH = None
        from nlp_tools.layers.normal_attention import NormalAttentionLayer as AttLayer
        sentence_input = Input(shape=(MAX_SENT_LENGTH,), dtype='int32')
        embedded_sequences = embed_model(sentence_input)
        l_lstm = Bidirectional(GRU(100, return_sequences=True))(embedded_sequences)
        l_att = AttLayer(100)(l_lstm)
        sentEncoder = Model(sentence_input, l_att)

        review_input = Input(shape=(20, MAX_SENT_LENGTH), dtype='int32')
        review_encoder = TimeDistributed(sentEncoder)(review_input)

        # review_encoder = HierarchicalAttentionMaskingLayer(input_masking=review_input)(review_encoder)
        l_lstm_sent = Bidirectional(GRU(100, return_sequences=True))(review_encoder)
        l_att_sent = AttLayer(100)(l_lstm_sent)
        preds = Dense(output_dim, activation='softmax')(l_att_sent)
        #model = Model(review_input, preds)

        self.tf_model: keras.Model = keras.Model(review_input, preds)


    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            from nlp_tools.loss import multi_category_focal_loss2_fixed
            from tensorflow.keras.losses import CategoricalCrossentropy
            loss = 'categorical_crossentropy'
            #loss = multi_category_focal_loss2_fixed



        optimizer = 'adam'
        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)



from nlp_tools.embeddings.abc_embedding import ABCEmbedding
from nlp_tools.embeddings.bare_embedding import BareEmbedding
from nlp_tools.embeddings.bert_embedding import BertEmbedding
from nlp_tools.embeddings.word_embedding import WordEmbedding
from nlp_tools.embeddings.transformer_embedding import TransformerEmbedding
from nlp_tools.layers import L
from nlp_tools.logger import  logger




# encoding: utf-8
from abc import ABC
from typing import Dict, List, Any, Optional, Union
import numpy as np
from tensorflow.keras.preprocessing.sequence import pad_sequences

from nlp_tools.generators import CorpusGenerator
from nlp_tools.processors.abc_sentence_processor import ABCSentenceProcessor
from nlp_tools.types import TextSamplesVar
from nlp_tools.utils import list_utils


class DialogueSequenceProcessor(ABCSentenceProcessor, ABC):
    """Generic processors for the sequence samples."""

    def to_dict(self) -> Dict[str, Any]:
        data = super(DialogueSequenceProcessor, self).to_dict()
        data['config'].update({
        })
        return data

    def __init__(self,
                 return_ner_masking=True,
                 **kwargs: Any) -> None:
        """

        Args:
            vocab_dict_type: initial vocab dict type, one of `text` `labeling`.
            **kwargs:
        """
        self.return_ner_masking = return_ner_masking
        super(DialogueSequenceProcessor, self).__init__(**kwargs)

    def update_length_info(self, embedding_max_position=512, max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length

    def transform(self, samples: TextSamplesVar, seq_length: int = None):
        ner_masking = []

        clean_seqs = []
        for seq in samples:
            sub_ner_masking = []
            start_position = 0
            new_seq = []
            for index, sub_seq in enumerate(seq):
                if "$$" in sub_seq:
                    split_tokens = sub_seq.split("$$")
                    sub_seq = sub_seq.replace("$$","")
                    pre_tokenized_seq = []
                    for item in split_tokens:
                        pre_tokenized_seq.extend(self.text_tokenizer.tokenize(item))
                    tokenized_seq = self.text_tokenizer.tokenize(sub_seq)

                    # 如果两次分词结果不一样，说明$$之间的关键词有问题，直接pass(因为关键词是直接基于字符匹配的，可能匹配到部分)
                    if not list_utils.check_list_same(pre_tokenized_seq, tokenized_seq):
                        sub_ner_masking.append(0)
                    else:
                        start_entity = len(self.text_tokenizer.tokenize(split_tokens[0])) + start_position
                        if index == 0:
                            # 如果是第一个位置，则要加上[cls]这个特殊词的位置
                            start_entity += 1

                        for i in range(start_entity, start_entity + len(self.text_tokenizer.tokenize(split_tokens[1]))):
                            sub_ner_masking.append(i)

                tokenized_seq = self.text_tokenizer.tokenize(sub_seq, max_len=seq_length, add_special_tokens=True)
                if index != 0:
                    tokenized_seq = tokenized_seq[1:]

                start_position += len(tokenized_seq)
                new_seq.extend(tokenized_seq)
            ner_masking.append(sub_ner_masking)
            clean_seqs.append(new_seq)

        tokenizers_result = self.text_tokenizer.encode(clean_seqs, maxlen=seq_length,is_split_into_words=True,add_special_tokens=False)

        ner_masking_ids = pad_sequences(ner_masking, max([len(i) for i in ner_masking]), padding='post',
                                        truncating='post')
        ner_masking_ids = np.array(ner_masking_ids)
        if self.return_ner_masking:
            return tokenizers_result + [ner_masking_ids]
        else:
            return tokenizers_result

    def inverse_transform(self,
                          labels: Union[List[List[int]], np.ndarray],
                          *,
                          lengths: List[int] = None,
                          threshold: float = 0.5,
                          **kwargs: Any) -> List[List[str]]:
        result = []
        for index, seq in enumerate(labels):
            labels_ = []
            for idx in seq:
                labels_.append(self.text_tokenizer.id_to_token(idx))
            if lengths is not None:
                labels_ = labels_[1:lengths[index] + 1]
            else:
                labels_ = labels_[1:-1]
            result.append(labels_)
        return result

    def get_max_seq_length_from_corpus(self,
                                       generators: List[CorpusGenerator],
                                       cover_rate: float = 0.95) -> int:
        """
        Calculate proper sequence length according to the corpus

        Args:
            generators:
            cover_rate:

        Returns:

        """
        return 512


if __name__ == "__main__":
    pass



from typing import Dict,Any
from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.simalary.abc_model import ABCSimilaryModel
from nlp_tools.layers.similary import LocalInferenceLayer




class EsimModel(ABCSimilaryModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            "first_bilstm":{
                'units':128,
                'return_sequences':True,
                'dropout':0.3
            },
            'composition_left':{
                'units':128,
                'return_sequences':True,
                'dropout':0.3
            },
            'composition_right': {
                'units': 128,
                'return_sequences': True,
                'dropout': 0.3
            },
            'mixed_dropout':{
                'rate':0.3
            },
            'mlp_layer': {
                "units": 128,
                'activation': 'relu'
            },
            "mlp_dropout": {
                'rate': 0.3
            }
    }





    def _make_inputs(self,seq_length) -> list:
        input_left = keras.layers.Input(
            name='text_left',
            shape=(seq_length,)
        )
        input_right = keras.layers.Input(
            name='text_right',
            shape=(seq_length,)
        )
        return [input_left, input_right]

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters
        embedding_model = self.embedding.embed_model

        first_bilstm_layer = L.Bidirectional(L.LSTM(**config["first_bilstm"]))
        composition_bilstm_left_layer = L.Bidirectional(L.LSTM(**config["composition_left"]))
        composition_bilstm_right_layer = L.Bidirectional(L.LSTM(**config["composition_right"]))
        mix_dropout_layer = L.Dropout(**config["mixed_dropout"])
        mlp_layer = L.Dense(**config['mlp_layer'])
        mlp_dropout = L.Dropout(**config['mlp_dropout'])


        # Left input and right input.
        input_left, input_right = self._make_inputs(self.sequence_length)
        input_left_embed ,input_right_embed = embedding_model(input_left),embedding_model(input_right)

        encoded_left = first_bilstm_layer(input_left_embed)
        encoded_right = first_bilstm_layer(input_right_embed)

        # ---------- Local inference layer ---------- #
        local_left,local_right = LocalInferenceLayer()([encoded_left,encoded_right])

        # ---------- Inference composition layer ---------- #
        composition_left = composition_bilstm_left_layer(local_left)
        composition_right = composition_bilstm_right_layer(local_right)

        avg_pool_left = L.GlobalAveragePooling1D()(composition_left)
        max_pool_left = L.GlobalMaxPooling1D()(composition_left)

        avg_pool_right = L.GlobalAveragePooling1D()(composition_right)
        max_pool_right = L.GlobalMaxPooling1D()(composition_right)

        mixed = L.concatenate([avg_pool_left,max_pool_left,avg_pool_right,max_pool_right])
        mixed = mix_dropout_layer(mixed)

        # ---------- Classification layer ---------- #
        mlp_result = mlp_layer(mixed)
        mlp_result = mlp_dropout(mlp_result)

        output = self.make_output_layer()(mlp_result)

        self.tf_model = keras.Model([input_left,input_right],output)


    def compile_model(self,
                      loss:Any = None,
                      optimizer: Any=None,
                      metrics:Any=None,
                      **kwargs:Any) -> None:
        if loss is None:
            loss = 'sparse_categorical_crossentropy'

        if optimizer is None:
            optimizer = 'adam'

        if metrics is None:
            metrics=["cosine_similarity"]

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)






from pretraining.pretraining import *
from pretraining.pretrain_data_utils import *

prtrained_model_type = 'roberta'

def load_training_tfrecord(pretrained_model_type,tf_corpus_paths,batch_size,grad_accum_steps,sequence_length):
    if pretrained_model_type == 'roberta':
        dataset = TrainingDatasetRoBERTa.load_tfrecord(
            record_names=tf_corpus_paths,
            sequence_length=sequence_length,
            batch_size=batch_size // grad_accum_steps,
        )

    elif pretrained_model_type == 'gpt':
        dataset = TrainingDatasetGPT.load_tfrecord(
            record_names=tf_corpus_paths,
            sequence_length=sequence_length,
            batch_size=batch_size // grad_accum_steps,
        )

    elif pretrained_model_type == 'unilm':

        dataset = TrainingDatasetUniLM.load_tfrecord(
            record_names=tf_corpus_paths,
            sequence_length=sequence_length,
            batch_size=batch_size // grad_accum_steps,
            token_sep_id=3,  # 这里需要自己指定[SEP]的id
        )
    else:
        raise ValueError("Not support")
    return dataset


def start_pretrained(
        model_saved_path,
        tf_record_path,
        bert_config_path,
        bert_checkpoint_path,# 如果从零训练，就设为None
        batch_size = 256,
        pretrained_model_type = 'roberta',
        tpu_address = None

):
    '''

    '''
    corpus_paths = []
    for file in os.listdir(tf_record_path):
        corpus_paths.append(os.path.join(tf_record_path, file))

    # 其他配置7
    sequence_length = 512
    learning_rate = 0.00176
    weight_decay_rate = 0.01
    num_warmup_steps = 3125
    num_train_steps = 125000
    steps_per_epoch = 10000
    grad_accum_steps = 16  # 大于1即表明使用梯度累积
    epochs = num_train_steps * grad_accum_steps // steps_per_epoch
    exclude_from_weight_decay = ['Norm', 'bias']
    exclude_from_layer_adaptation = ['Norm', 'bias']
    which_optimizer = 'lamb'  # adam 或 lamb，均自带weight decay
    lr_schedule = {
        num_warmup_steps * grad_accum_steps: 1.0,
        num_train_steps * grad_accum_steps: 0.0,
    }
    floatx = K.floatx()

    dataset = load_training_tfrecord(pretrained_model_type,corpus_paths,batch_size,grad_accum_steps,sequence_length)

    if tpu_address is None:
        # 单机多卡模式（多机多卡也类似，但需要硬软件配合，请参考https://tf.wiki）
        strategy = tf.distribute.MirroredStrategy()
    else:
        # TPU模式
        resolver = tf.distribute.cluster_resolver.TPUClusterResolver(
            tpu=tpu_address
        )
        tf.config.experimental_connect_to_host(resolver.master())
        tf.tpu.experimental.initialize_tpu_system(resolver)
        strategy = tf.distribute.experimental.TPUStrategy(resolver)

    with strategy.scope():
        train_model,bert_instance = build_transformer_model_for_pretraining(
            config_path=bert_config_path,
            data_dtype=floatx,
            which_optimizer=which_optimizer,
            learning_rate=learning_rate,
            lr_schedule=lr_schedule,
            weight_decay_rate=weight_decay_rate,
            exclude_from_weight_decay=exclude_from_weight_decay,
            exclude_from_layer_adaptation=exclude_from_layer_adaptation,
            grad_accum_steps=grad_accum_steps,
            checkpoint_path=bert_checkpoint_path,
            pretrained_model_type=pretrained_model_type
        )
        train_model.summary()

    # 保存模型
    checkpoint = ModelCheckpoint(model_saved_path=model_saved_path,bert_instance=bert_instance)
    # 记录日志
    csv_logger = keras.callbacks.CSVLogger('training.log')

    # 模型训练
    train_model.fit(
        dataset,
        steps_per_epoch=steps_per_epoch,
        epochs=epochs,
        callbacks=[checkpoint, csv_logger],
    )


from typing import Dict, Any

from tensorflow import keras
from nlp_tools.layers import L
from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel
from nlp_tools.layers.pool_layer import PoolerInputLayer
from nlp_tools.layers.non_masking_layer import MaskingSaverLayer
import tensorflow as tf




class SpanClassifyModel(ABCLabelingModel):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_blstm': {
                'units': 64,
                'return_sequences': True
            },
            'layer_dropout': {
                'rate': 0.5
            },
            'layer_time_distributed': {},
        }



    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model
        self.mask_save_layer = MaskingSaverLayer()



        layers_starts_label_fc = L.Dense(output_dim)
        layers_ends_label_fc = PoolerInputLayer(128,output_dim)


        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_blstm']), name='layer_blstm'),
            L.Dropout(**config['layer_dropout'], name='layer_dropout'),
        ]

        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        tensor = self.mask_save_layer(tensor)

        starts_logits =  layers_starts_label_fc(tensor)
        ends_logits = layers_ends_label_fc([tensor,starts_logits])

        output_starts = tf.nn.softmax(starts_logits)
        output_ends = tf.nn.softmax(ends_logits)





        self.tf_model = keras.Model(embed_model.inputs, [output_starts,output_ends])


    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            self.mask_save_layer.set_loss_func(tf.keras.losses.sparse_categorical_crossentropy)
            loss = self.mask_save_layer.loss
        if metrics is None:
            metrics = 'accuracy'
        super(SpanClassifyModel, self).compile_model(loss=loss,
                                                    optimizer=optimizer,
                                                    metrics=metrics,
                                                    **kwargs)


    def predict(self,
                x_data,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs: Dict = None) :

        if predict_kwargs is None:
            predict_kwargs = {}

        if truncating:
            seq_length = self.max_sequence_length
        else:
            seq_length = None

        if type(x_data[0]) == list:
            x_data = ["".join(x).replace("##","").replace('[CLS]',"").replace("[SEP]","") for x in x_data]
        x_data_tokenized = [self.text_processor.text_tokenizer.tokenize(x) for x in x_data]
        lengths = [len(x) for x in x_data_tokenized]

        tensor = self.text_processor.transform(x_data,seq_length=seq_length)
        (starts_labels,ends_labels) = self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)
        #pred= self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)
        starts_labels = starts_labels.argmax(-1)
        ends_labels = ends_labels.argmax(-1)

        x_data_mapping = [self.text_processor.text_tokenizer.rematch(x, x_tokens) for x, x_tokens in
                          zip(x_data, x_data_tokenized)]

        res = self.label_processor.inverse_transform((starts_labels.tolist(),ends_labels.tolist()),lengths=lengths,mapping_list=x_data_mapping)
        return res




from math import log2,floor
from typing import Dict,Any
import tensorflow as tf

from nlp_tools.layers import L,KMaxPoolingLayer
from nlp_tools.tasks.classification.abc_model import ABCClassificationModel

class DPCNN_Model(ABCClassificationModel):
    '''
        This implementation of DPCNN requires a clear declared sequence length.
        So sequences input in should be padded or cut to a given length in advance.
        '''
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str,Dict[str,Any]]:
        pool_type = 'max'
        filters = 128
        activation = 'linear'

        return {
            'region_embedding':{
                'filters':filters,
                'kernel_size':3,
                'strides':1,
                'padding':"same",
                'activation':activation,
                'name':'region_embedding',
            },
            'region_dropout':{
                'rate':0.2,
            },
            'conv_block':{
                'fliters':filters,
                'kernel_size':3,
                'activation':activation,
                'shortcut':True,
            },
            'resnet_block':{
                'filters':filters,
                'kernel_size':3,
                'activation':activation,
                'shortcut':True,
                'pool_type':pool_type,
                'sorted':True
            },
            'dense':{
                'units':256,
                'activation':activation
            },
            'dropout':{
                'rate':0.5,
            },
            'activation':{
                'activation':"softmax"
            }
        }

    def downsample(self,inputs,pool_type:str='max',
                   sorted:bool=True,stage:int=1):
        layers_pool = []
        if pool_type == 'max':
            layers_pool.append(
                L.MaxPooling1D(pool_size=3,
                               strides=2,
                               padding='same',
                               name=f'pool_{stage}')
            )
        elif pool_type == 'k_max':
            k = int(inputs.shape[1].value / 2)
            layers_pool.append(
                KMaxPoolingLayer(k=k,
                                 sorted =sorted,
                                 name=f'pool_{stage}')

            )
        elif pool_type  == 'conv':
            layers_pool.append(
                L.Conv1D(filters=inputs.shape[-1].value,
                         kernel_size=3,
                         strides=2,
                         padding='same',
                         name=f'pool_{stage}')
            )
        elif pool_type is None:
            layers_pool = []
        else:
            raise ValueError("unsupported poolingtype `{pool_type}`~")

        tensor_out = inputs
        for layer in layers_pool:
            tensor_out = layer(tensor_out)
        return tensor_out

    def conv_block(self,inputs,filters,kernel_size:int=3,
                   activation:str='linear',shortcut:bool=True):
        layers_conv_unit = []
        layers_conv_unit.append(L.BatchNormalization())
        layers_conv_unit.append(L.PReLU())
        layers_conv_unit.append(
            L.Conv1D(filters=filters,
                     kernel_size=kernel_size,
                     strides=1,
                     padding='same',
                     activation=activation)
        )

        layers_conv_block = layers_conv_unit  * 2

        tensor_out = inputs
        for layer in layers_conv_block:
            tensor_out = layer(tensor_out)

        if shortcut:
            tensor_out = L.Add()([inputs,tensor_out])

        return tensor_out

    def resnet_block(self,inputs,filters:int,kernel_size:int=3,
                     activation:str='linear',shortcut:bool=True,
                     pool_type:str='max',sorted:bool=True,stage:int=1):
        tensor_pool = self.downsample(inputs,pool_type=pool_type,sorted=sorted,stage=stage)
        tensor_out = self.conv_block(tensor_pool,filters=filters,kernel_size=kernel_size,
                                     activation=activation,shortcut=shortcut)
        return tensor_out

    def build_model_arc(self):
        output_dim = len(self.label_processor.vocab2idx)
        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layers_region = [
            L.Conv1D(**config['region_embedding']),
            L.BatchNormalization(),
            L.PReLU(),
            L.Dropout(**config['region_dropout'])

        ]

        layers_main = [
           L.GlobalMaxPooling1D(),
           L.Dense(**config['dense']),
           L.BatchNormalization(),
           L.PReLU(),
           L.Dropout(**config['dropout']),
           L.Dense(output_dim,**config['activation'])
        ]

        tensor_out = embed_model.output
        tensor_out = tf.reshape(tensor_out,(-1,self.max_sequence_length,tensor_out.shape[-1]))

        for layer in layers_region:
            tensor_out = layer(tensor_out)

        tensor_out = self.conv_block(tensor_out,filters=config['conv_block']['fliters'],
                                        kernel_size= config['conv_block']['kernel_size'],
                                        activation = config['conv_block']['activation'],
                                        shortcut=config['conv_block']['shortcut'])

        seq_len = tensor_out.shape[1]
        if seq_len is None:
            raise ValueError('sequence_length should be explicitly assigned,but it is None')

        for layer in layers_main:
            tensor_out = layer(tensor_out)

        self.tf_model = tf.keras.Model(embed_model.inputs,tensor_out)










import pandas as pd
from nlp_tools import utils

class DataFoundClassify():
    '''
    第五届“达观杯” 基于大规模预训练模型的风险事件标签识别
    https://www.datafountain.cn/competitions/512
    '''

    @classmethod
    def load_data(cls,
                  file_name: str ,
                  shuffle: bool = True):
        data_df = pd.read_csv(file_name)
        x_data = data_df['text'].tolist()
        y_data = data_df['label'].tolist()

        
        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)

        data_union = [(x, y) for x, y in zip(x_data, y_data)]
        return data_union


if __name__ == '__main__':
    data_path = r'/home/qiufengfeng/nlp/competition/datagrand/基于大规模预训练模型的风险事件标签识别/processed/train.csv'

    DataFoundClassify().load_data(file_name=data_path)


import os
from typing import Dict,Any
os.environ['TF_KERAS'] = '1'
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

custom_objects: Dict[str, Any] = {}

from nlp_tools.__version__ import __version__
from nlp_tools.macros import config
from nlp_tools import layers
from nlp_tools import corpus
from nlp_tools import embeddings
from nlp_tools import macros
from nlp_tools import processors
from nlp_tools import tasks
from nlp_tools import utils
#from nlp_tools import callbacks

custom_objects = layers.resigter_custom_layers(custom_objects)
from tensorflow import keras
keras.utils.get_custom_objects().update(custom_objects)


# encoding: utf-8

# author: BrikerMan
# contact: eliyar917@gmail.com
# blog: https://eliyar.biz

# file: abc_model.py
# time: 4:30 下午

from abc import ABC
from typing import List, Dict, Any, Union, Optional

import numpy as np
import tensorflow as tf

import nlp_tools
from nlp_tools.embeddings import ABCEmbedding, BareEmbedding,TransformerEmbedding,BertEmbedding
from nlp_tools.generators import CorpusGenerator, BatchGenerator
from nlp_tools.logger import logger
from nlp_tools.metrics.sequence_labeling import get_entities
from nlp_tools.metrics.sequence_labeling import sequence_labeling_report
from nlp_tools.processors import SequenceProcessor
from nlp_tools.tasks.abs_task_model import ABCTaskModel
from nlp_tools.types import TextSamplesVar
from nlp_tools.optimizer import MultiOptimizer

from tensorflow.keras.optimizers import Adam
from nlp_tools.utils.ner_utils import output_ner_results
from nlp_tools.utils.training_tips import creat_FGM
from nlp_tools.loss.r_drop_loss import RDropLoss
from sklearn.model_selection._split import StratifiedKFold,KFold
from tensorflow.keras import backend as K
import functools
import os



class MutitaskClassifyAndNer(ABCTaskModel, ABC):
    """
    Abstract Labeling Model
    """

    def __init__(self,
                 **kwargs):
        """

        Args:
            embedding: embedding object
            max_sequence_length: target sequence length
            hyper_parameters: hyper_parameters to overwrite
        """
        super(MutitaskClassifyAndNer, self).__init__(**kwargs)





    def build_model_arc(self) -> None:
        raise NotImplementedError

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            loss = 'sparse_categorical_crossentropy'


        if type(self.embedding) in [TransformerEmbedding,BertEmbedding] :
            total_layers = self.tf_model.layers
            transfomer_layers = self.embedding.embed_model.layers
            no_transformer_layers = [layer for layer in total_layers if layer not in transfomer_layers]
            optimizer_list = [
                Adam(learning_rate=1e-5),
                Adam(learning_rate=2e-5)
            ]
            optimizers_and_layers = [(optimizer_list[0], no_transformer_layers), (optimizer_list[1], transfomer_layers)]
            optimizer = MultiOptimizer(optimizers_and_layers)




        if optimizer is None:
            optimizer = Adam(learning_rate=1e-5)
        if self.use_rdrop:
            if type(loss) == list:
                loss = [RDropLoss(i) for i in loss]
            else:
                loss = RDropLoss(loss)


        if metrics is None:
            metrics = ['accuracy']

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)





from abc import ABC
from typing import Dict,List,Any,Tuple
import numpy as np

from nlp_tools.generators import  CorpusGenerator
from nlp_tools.types import TextSamplesVar
from nlp_tools.processors.abc_label_processor import ABCLabelProcessor



class MutiProcessors(ABCLabelProcessor):
    def to_dict(self) -> Dict[str, Any]:
        processed_dict_list = []
        for processor in self.processors:
            processed_dict_list.append(processor.to_dict())

        return {
            'config': {
                'processors': processed_dict_list
            },
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
        }

    def __init__(self,processors) -> None:
        self.processors = processors



    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        for processor in self.processors:
            processor.build_vocab_generator(generators)

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        raise NotImplementedError

    def transform(self,samples,seq_length = None):
        output = []
        for processor in self.processors:
            transform_data = processor.transform(seq_length)
            if type(transform_data) == tuple or type(transform_data) == list:
                output.extend(list(transform_data))
            else:
                output.append(transform_data)
        return output

    def inverse_transform(self,
                          labels,
                          lengths = None,
                          mapping_list=None,
                          threshold= 0.5) :
        raise NotImplementedError

    def _override_load_model(self,config:Dict) -> None:
        self.processors = []
        for processor_dict  in  config['processors']:
            self.processors.append(load_data_object(processor_dict))

    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length






from typing import Tuple, List
from nlp_tools import macros as k
import os
from nlp_tools.corpus import DataReader
from nlp_tools import utils
import logging
import json
### txt sample line:
### {"summarization": "知情人透露章子怡怀孕后，父母很高兴。章母已开始悉心照料。据悉，预产期大概是12月底", "article": "四海网讯，近日，有媒体报道称：章子怡真怀孕了!报道还援引知情人士消息称，“章子怡怀孕大概四五个月，预产期是年底前后，现在已经不接工作了。”这到底是怎么回事?消息是真是假?针对此消息，23日晚8时30分，华西都市报记者迅速联系上了与章子怡家里关系极好的知情人士，这位人士向华西都市报记者证实说：“子怡这次确实怀孕了。她已经36岁了，也该怀孕了。章子怡怀上汪峰的孩子后，子怡的父母亲十分高兴。子怡的母亲，已开始悉心照料女儿了。子怡的预产期大概是今年12月底。”当晚9时，华西都市报记者为了求证章子怡怀孕消息，又电话联系章子怡的亲哥哥章子男，但电话通了，一直没有人<Paragraph>接听。有关章子怡怀孕的新闻自从2013年9月份章子怡和汪峰恋情以来，就被传N遍了!不过，时间跨入2015年，事情却发生着微妙的变化。2015年3月21日，章子怡担任制片人的电影《从天儿降》开机，在开机发布会上几张合影，让网友又燃起了好奇心：“章子怡真的怀孕了吗?”但后据证实，章子怡的“大肚照”只是影片宣传的噱头。过了四个月的7月22日，《太平轮》新一轮宣传，章子怡又被发现状态不佳，不时深呼吸，不自觉想捂住肚子，又觉得不妥。然后在8月的一天，章子怡和朋友吃饭，在酒店门口被风行工作室拍到了，疑似有孕在身!今年7月11日，汪峰本来在上海要举行演唱会，后来因为台风“灿鸿”取消了。而消息人士称，汪峰原来打算在演唱会上当着章子怡的面宣布重大消息，而且章子怡已经赴上海准备参加演唱会了，怎知遇到台风，只好延期，相信9月26日的演唱会应该还会有惊喜大白天下吧。"}



class NlpCC2017TextSummarization(object):
    __corpus_name__ = 'nlp_cc2017_text_summarization'

    @classmethod
    def load_data(cls,
                  file_path: str ,
                  shuffle: bool = True) -> Tuple[List[List[str]], List[List[str]]]:
        """
                Load dataset as sequence labeling format, char level tokenized

                features: ``[['海', '钓', '比', '赛', '地', '点', '在', '厦', '门', ...], ...]``

                labels: ``[[海', '钓', '比', '赛', '地', '点', , ...], ...]``

                Args:
                    file_path: file ablsute path
                    shuffle: should shuffle or not, default True.

                Returns:
                    dataset_features and dataset labels
                """
        x_data = []
        y_data = []
        with open(file_path,'r',encoding='utf-8') as fread:
            for index, line in enumerate(fread):
                json_obj = json.loads(line.strip())
                x_data.append(list(json_obj['summarization']))
                y_data.append(list(json_obj['article']))

        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)
        logging.debug(f"loaded {len(x_data)} samples from {file_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")
        return x_data, y_data


import re
def _regular_(sen):
    """
    句子规范化，主要是对原始语料的句子进行一些标点符号的统一
    :param sen:
    :return:
    """
    sen = sen.replace('/', '')
    sen = re.sub(r'…{1,100}', '…', sen)
    sen = re.sub(r'\.{3,100}', '…', sen)
    sen = re.sub(r'···{2,100}', '…', sen)
    sen = re.sub(r',{1,100}', '，', sen)
    sen = re.sub(r'\.{1,100}', '。', sen)
    sen = re.sub(r'。{1,100}', '。', sen)
    sen = re.sub(r'\?{1,100}', '？', sen)
    sen = re.sub(r'？{1,100}', '？', sen)
    sen = re.sub(r'!{1,100}', '！', sen)
    sen = re.sub(r'！{1,100}', '！', sen)
    sen = re.sub(r'~{1,100}', '～', sen)
    sen = re.sub(r'～{1,100}', '～', sen)
    sen = re.sub(r'[“”]{1,100}', '"', sen)
    sen = re.sub('[^\w\u4e00-\u9fff"。，？！～·]+', '', sen)
    sen = re.sub(r'[ˇˊˋˍεπのゞェーω]', '', sen)

    return sen


def full_to_half(s):
    """
    Convert full-width character to half-width one
    """
    n = []
    for char in s:
        num = ord(char)
        if num == 0x3000:
            num = 32
        elif 0xFF01 <= num <= 0xFF5E:
            num -= 0xfee0
        char = chr(num)
        n.append(char)
    return ''.join(n)


def cut_to_sentence(text):
    """
    Cut text to sentences
    """
    sentence = []
    sentences = []
    len_p = len(text)
    pre_cut = False
    for idx, word in enumerate(text):
        sentence.append(word)
        cut = False
        if pre_cut:
            cut=True
            pre_cut=False
        if word in u"。;!?\n":
            cut = True
            if len_p > idx+1:
                if text[idx+1] in ".。”\"\'“”‘’?!":
                    cut = False
                    pre_cut=True

        if cut:
            sentences.append(sentence)
            sentence = []
    if sentence:
        sentences.append("".join(list(sentence)))
    return sentences


def replace_html(s):
    s = s.replace('&quot;','"')
    s = s.replace('&amp;','&')
    s = s.replace('&lt;','<')
    s = s.replace('&gt;','>')
    s = s.replace('&nbsp;',' ')
    s = s.replace("&ldquo;", "“")
    s = s.replace("&rdquo;", "”")
    s = s.replace("&mdash;","")
    s = s.replace("\xa0", " ")
    return(s)


from typing import Dict, Any

from tensorflow import keras
from tensorflow.keras import backend as K
from nlp_tools.layers import L, KConditionalRandomField
from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel
from tensorflow.keras.preprocessing.sequence import  pad_sequences




class KfoldCrfModel(ABCLabelingModel):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {}


    def build_model_arc(self) -> None:
        label_size = self.label_processor.vocab_size

        crf = KConditionalRandomField()
        input = L.Input(shape=(None,label_size),dtype=K.floatx())
        input_masking = L.Input(shape=(None,))



        tensor = crf(input,input_masking)

        self.tf_model = keras.Model([input,input_masking], tensor)
        self.crf_layer = crf

    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:

        if loss is None:
            loss = self.crf_layer.loss
        if metrics is None:
            metrics = [self.crf_layer.accuracy]
        super(KfoldCrfModel, self).compile_model(loss=loss,
                                                    optimizer=optimizer,
                                                    metrics=metrics,
                                                    **kwargs)


    def predict(self,
                x_data,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs=None,
                return_max_entity=False) :
        if predict_kwargs is None:
            predict_kwargs = {}

        if truncating:
            seq_length = self.max_sequence_length
        else:
            seq_length = None

        sentences = []
        y_folds = []
        for item in x_data:
            sentences.append(item[0])
            y_folds.append(item[1])



        x_data_tokenized = [list(x) for x in sentences]
        lengths = [len(x) for x in x_data_tokenized]

        max_sequence_len = max(lengths)
        masking = []
        for seq in sentences:
            seq_masking = [1] * len(seq)
            masking.append(seq_masking)
        masking = pad_sequences(masking, max_sequence_len, padding='post', truncating='post')

        y_folds = self.label_processor.transform_kfold(y_folds, seq_length=max_sequence_len)



        pred = self.tf_model.predict((y_folds,masking), batch_size=batch_size, **predict_kwargs)
        pred = pred.argmax(-1)

        res = self.label_processor.inverse_transform(pred,lengths=lengths,mapping_list =None)
        return res


    def predict_entities(self,
                         x_data,
                         batch_size: int = 32,
                         truncating: bool = False,
                         predict_kwargs: Dict = None) :
        predict_result = self.predict(x_data,batch_size,truncating,predict_kwargs)

        return predict_result




#! -*- coding: utf-8 -*-
# 预训练语料构建

import os
os.environ['TF_KERAS'] = '1'  # 必须使用tf.keras

import numpy as np
import tensorflow as tf
from bert4keras.snippets import parallel_apply
from bert4keras.backend import K


class TrainingDataset(object):
    """预训练数据集生成器
    """
    def __init__(self, tokenizer, sequence_length=512):
        """参数说明：
            tokenizer必须是bert4keras自带的tokenizer类；
        """
        self.tokenizer = tokenizer
        self.sequence_length = sequence_length
        self.token_pad_id = tokenizer._token_pad_id
        self.token_cls_id = tokenizer._token_start_id
        self.token_sep_id = tokenizer._token_end_id
        self.token_mask_id = tokenizer._token_mask_id
        self.vocab_size = tokenizer._vocab_size

    def padding(self, sequence, padding_value=None):
        """对单个序列进行补0
        """
        if padding_value is None:
            padding_value = self.token_pad_id

        sequence = sequence[:self.sequence_length]
        padding_length = self.sequence_length - len(sequence)
        return sequence + [padding_value] * padding_length

    def sentence_process(self, text):
        """单个文本的处理函数，返回处理后的instance
        """
        raise NotImplementedError

    def paragraph_process(self, texts, starts, ends, paddings):
        """单个段落（多个文本）的处理函数
        说明：texts是单句组成的list；starts是每个instance的起始id；
              ends是每个instance的终止id；paddings是每个instance的填充id。
        做法：不断塞句子，直到长度最接近sequence_length，然后padding。
        """
        instances, instance = [], [[start] for start in starts]

        # for text in texts:
        #     # 处理单个句子
        sub_instance = self.sentence_process(texts)
        sub_instance = [i[:self.sequence_length - 2] for i in sub_instance]
        new_length = len(instance[0]) + len(sub_instance[0])

        # 如果长度即将溢出
        if new_length > self.sequence_length - 1:
            # 插入终止符，并padding
            complete_instance = []
            for item, end, pad in zip(instance, ends, paddings):
                item.append(end)
                item = self.padding(item, pad)
                complete_instance.append(item)
            # 存储结果，并构建新样本
            instances.append(complete_instance)
            instance = [[start] for start in starts]

        # 样本续接
        for item, sub_item in zip(instance, sub_instance):
            item.extend(sub_item)

        # 插入终止符，并padding
        complete_instance = []
        for item, end, pad in zip(instance, ends, paddings):
            item.append(end)
            item = self.padding(item, pad)
            complete_instance.append(item)

        # 存储最后的instance
        instances.append(complete_instance)

        return instances

    def tfrecord_serialize(self, instances, instance_keys):
        """转为tfrecord的字符串，等待写入到文件
        """
        def create_feature(x):
            return tf.train.Feature(int64_list=tf.train.Int64List(value=x))

        serialized_instances = []
        for instance in instances:
            features = {
                k: create_feature(v)
                for k, v in zip(instance_keys, instance)
            }
            tf_features = tf.train.Features(feature=features)
            tf_example = tf.train.Example(features=tf_features)
            serialized_instance = tf_example.SerializeToString()
            serialized_instances.append(serialized_instance)

        return serialized_instances

    def process(self, corpus, record_name, workers=8, max_queue_size=2000):
        """处理输入语料（corpus），最终转为tfrecord格式（record_name）
        自带多进程支持，如果cpu核心数多，请加大workers和max_queue_size。
        """
        writer = tf.io.TFRecordWriter(record_name)
        globals()['count'] = 0

        def write_to_tfrecord(serialized_instances):
            globals()['count'] += len(serialized_instances)
            for serialized_instance in serialized_instances:
                writer.write(serialized_instance)

        def paragraph_process(texts):
            instances = self.paragraph_process(texts)
            serialized_instances = self.tfrecord_serialize(instances)
            return serialized_instances

        parallel_apply(
            func=paragraph_process,
            iterable=corpus,
            workers=workers,
            max_queue_size=max_queue_size,
            callback=write_to_tfrecord,
        )

        writer.close()
        print('write %s examples into %s' % (count, record_name))

    @staticmethod
    def load_tfrecord(record_names, batch_size, parse_function):
        """加载处理成tfrecord格式的语料
        """
        if not isinstance(record_names, list):
            record_names = [record_names]

        dataset = tf.data.TFRecordDataset(record_names)  # 加载
        dataset = dataset.map(parse_function)  # 解析
        dataset = dataset.repeat()  # 循环
        dataset = dataset.shuffle(batch_size * 1000)  # 打乱
        dataset = dataset.batch(batch_size)  # 成批

        return dataset


class TrainingDatasetRoBERTa(TrainingDataset):
    """预训练数据集生成器（RoBERTa模式）
    """
    def __init__(
        self, tokenizer, word_segment, mask_rate=0.15, sequence_length=512
    ):
        """参数说明：
            tokenizer必须是bert4keras自带的tokenizer类；
            word_segment是任意分词函数。
        """
        super(TrainingDatasetRoBERTa, self).__init__(tokenizer, sequence_length)
        self.word_segment = word_segment
        self.mask_rate = mask_rate

    def token_process(self, token_id):
        """以80%的几率替换为[MASK]，以10%的几率保持不变，
        以10%的几率替换为一个随机token。
        """
        rand = np.random.random()
        if rand <= 0.8:
            return self.token_mask_id
        elif rand <= 0.9:
            return token_id
        else:
            return np.random.randint(0, self.vocab_size)

    def sentence_process(self, text):
        """单个文本的处理函数
        流程：分词，然后转id，按照mask_rate构建全词mask的序列
              来指定哪些token是否要被mask
        """
        words = self.word_segment(text)
        rands = np.random.random(len(words))

        token_ids, mask_ids = [], []
        for rand, word in zip(rands, words):
            word_tokens = self.tokenizer.tokenize(text=word)[1:-1]
            word_token_ids = self.tokenizer.tokens_to_ids(word_tokens)
            token_ids.extend(word_token_ids)

            if rand < self.mask_rate:
                word_mask_ids = [
                    self.token_process(i) + 1 for i in word_token_ids
                ]
            else:
                word_mask_ids = [0] * len(word_tokens)

            mask_ids.extend(word_mask_ids)

        return [token_ids, mask_ids]

    def paragraph_process(self, texts):
        """给原方法补上starts、ends、paddings
        """
        starts = [self.token_cls_id, 0]
        ends = [self.token_sep_id, 0]
        paddings = [self.token_pad_id, 0]
        return super(TrainingDatasetRoBERTa,
                     self).paragraph_process(texts, starts, ends, paddings)

    def tfrecord_serialize(self, instances):
        """给原方法补上instance_keys
        """
        instance_keys = ['token_ids', 'mask_ids']
        return super(TrainingDatasetRoBERTa,
                     self).tfrecord_serialize(instances, instance_keys)

    @staticmethod
    def load_tfrecord(record_names, sequence_length, batch_size):
        """给原方法补上parse_function
        """
        def parse_function(serialized):
            features = {
                'token_ids': tf.io.FixedLenFeature([sequence_length], tf.int64),
                'mask_ids': tf.io.FixedLenFeature([sequence_length], tf.int64),
            }
            features = tf.io.parse_single_example(serialized, features)
            token_ids = features['token_ids']
            mask_ids = features['mask_ids']
            segment_ids = K.zeros_like(token_ids, dtype='int64')
            is_masked = K.not_equal(mask_ids, 0)
            masked_token_ids = K.switch(is_masked, mask_ids - 1, token_ids)
            x = {
                'Input-Token': masked_token_ids,
                'Input-Segment': segment_ids,
                'token_ids': token_ids,
                'is_masked': K.cast(is_masked, K.floatx()),
            }
            y = {
                'mlm_loss': K.zeros([1]),
                'mlm_acc': K.zeros([1]),
            }
            return x, y

        return TrainingDataset.load_tfrecord(
            record_names, batch_size, parse_function
        )


class TrainingDatasetGPT(TrainingDataset):
    """预训练数据集生成器（GPT模式，单向语言模型）
    """
    def sentence_process(self, text):
        """单个文本的处理函数
        流程：分词，然后转id。
        """
        tokens = self.tokenizer.tokenize(text=text)[1:-1]
        token_ids = self.tokenizer.tokens_to_ids(tokens)
        return [token_ids]

    def paragraph_process(self, texts):
        """给原方法补上starts、ends、paddings
        """
        starts = [self.token_cls_id]
        ends = [self.token_sep_id]
        paddings = [self.token_pad_id]
        return super(TrainingDatasetGPT,
                     self).paragraph_process(texts, starts, ends, paddings)

    def tfrecord_serialize(self, instances):
        """给原方法补上instance_keys
        """
        instance_keys = ['token_ids']
        return super(TrainingDatasetGPT,
                     self).tfrecord_serialize(instances, instance_keys)

    @staticmethod
    def load_tfrecord(record_names, sequence_length, batch_size):
        """给原方法补上parse_function
        """
        def parse_function(serialized):
            features = {
                'token_ids': tf.io.FixedLenFeature([sequence_length], tf.int64),
            }
            features = tf.io.parse_single_example(serialized, features)
            token_ids = features['token_ids']
            segment_ids = K.zeros_like(token_ids, dtype='int64')
            x = {
                'Input-Token': token_ids,
                'Input-Segment': segment_ids,
            }
            y = {
                'lm_loss': K.zeros([1]),
                'lm_acc': K.zeros([1]),
            }
            return x, y

        return TrainingDataset.load_tfrecord(
            record_names, batch_size, parse_function
        )


class TrainingDatasetUniLM(TrainingDatasetGPT):
    """预训练数据集生成器（UniLM模式，Seq2Seq模型）
    """
    @staticmethod
    def load_tfrecord(record_names, sequence_length, batch_size, token_sep_id):
        """给原方法补上parse_function
        """
        def parse_function(serialized):
            features = {
                'token_ids': tf.io.FixedLenFeature([sequence_length], tf.int64),
            }
            features = tf.io.parse_single_example(serialized, features)
            token_ids = features['token_ids']
            segment = K.random_uniform(
                shape=[1], minval=1, maxval=sequence_length - 1, dtype='int64'
            )[0]
            segment_ids = K.one_hot(segment + 1, sequence_length)
            segment_ids = K.cast(K.cumsum(segment_ids), 'int64')
            token_ids_1 = token_ids[:segment]
            token_ids_2 = K.zeros([1], dtype='int64') + token_sep_id
            token_ids_3 = token_ids[segment:-1]
            token_ids = K.concatenate([token_ids_1, token_ids_2, token_ids_3])
            x = {
                'Input-Token': token_ids,
                'Input-Segment': segment_ids,
            }
            y = {
                'unilm_loss': K.zeros([1]),
                'unilm_acc': K.zeros([1]),
            }
            return x, y

        return TrainingDataset.load_tfrecord(
            record_names, batch_size, parse_function
        )


if __name__ == '__main__':

    from bert4keras.tokenizers import Tokenizer
    import json, glob, re
    from tqdm import tqdm

    model = 'roberta'
    sequence_length = 512
    workers = 40
    max_queue_size = 4000
    dict_path = '/home/spaces_ac_cn/chinese_L-12_H-768_A-12/vocab.txt'
    tokenizer = Tokenizer(dict_path, do_lower_case=True)

    def some_texts():
        filenames = glob.glob('/home/spaces_ac_cn/corpus/*/*/*')
        np.random.shuffle(filenames)
        count, texts = 0, []
        for filename in filenames:
            with open(filename) as f:
                for l in f:
                    l = json.loads(l)['text'].strip()
                    texts.extend(re.findall(u'.*?[\n。]+', l))
                    count += 1
                    if count == 10:  # 10篇文章合在一起再处理
                        yield texts
                        count, texts = 0, []
        if texts:
            yield texts

    assert model in ['roberta', 'gpt', 'unilm']  # 判断是否支持的模型类型

    if model == 'roberta':

        import jieba_fast as jieba
        jieba.initialize()

        def word_segment(text):
            return jieba.lcut(text)

        TD = TrainingDatasetRoBERTa(
            tokenizer, word_segment, sequence_length=sequence_length
        )

        for i in range(10):  # 数据重复10遍
            TD.process(
                corpus=tqdm(some_texts()),
                record_name='../corpus_tfrecord/corpus.%s.tfrecord' % i,
                workers=workers,
                max_queue_size=max_queue_size,
            )

    elif model == 'gpt':

        TD = TrainingDatasetGPT(tokenizer, sequence_length=sequence_length)

        TD.process(
            corpus=tqdm(some_texts()),
            record_name='../corpus_tfrecord/corpus.tfrecord',
            workers=workers,
            max_queue_size=max_queue_size,
        )

    elif model == 'unilm':

        TD = TrainingDatasetUniLM(tokenizer, sequence_length=sequence_length)

        TD.process(
            corpus=tqdm(some_texts()),
            record_name='../corpus_tfrecord/corpus.tfrecord',
            workers=workers,
            max_queue_size=max_queue_size,
        )



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   AutoEmbeddingPrefixEmbedding.py    
@Contact :   544855237@qq.com

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/12/22 下午4:36   qiufengfeng      1.0         None
'''

# import lib
from transformers import TFAutoModel,AutoConfig,AutoTokenizer
import tensorflow as tf
from tensorflow.keras import layers,Model
import numpy as np

from nlp_tools.embeddings.abc_embedding import ABCEmbedding
from nlp_tools.embeddings.hugginface.prefix_prompt.configuration_prefix_prompt_roberta import PrefixPromptConfig
from nlp_tools.utils.prompt.prefix_encoder import get_prompt_values
from typing import Dict,Any


class AutoEmbeddingPrefixEmbedding(ABCEmbedding):
    def to_dict(self) -> Dict[str, Any]:
        info_dic = super(AutoEmbeddingPrefixEmbedding, self).to_dict()
        return info_dic

    def __init__(self,pretrained_model_path,inputs_keys,**kwargs):
        super(AutoEmbeddingPrefixEmbedding, self).__init__(**kwargs)

        self.pretrained_model_path = pretrained_model_path
        self.inputs_keys = inputs_keys
        self.max_position = 512


    def make_inputs_nodes(self):
        input_dict = {}
        for key in self.inputs_keys:
            input_dict[key] = layers.Input(shape=(None,),dtype=tf.int32)
        return input_dict






    def build_embedding_model(self) -> None:
        config = PrefixPromptConfig.from_pretrained(self.pretrained_model_path)
        inputs = self.make_inputs_nodes()
        input_list = [value for key,value in inputs.items()]

        batch_size = tf.shape(inputs['input_ids'])[0]  # 4
        past_key_values, prefix_attention_mask = get_prompt_values(config=config,batch_size=batch_size)

        autoBertModel = TFAutoModel.from_pretrained(self.pretrained_model_path)

        inputs['attention_mask'] = layers.concatenate((prefix_attention_mask, inputs['attention_mask']), axis=1)  # shape [4,136]
        inputs['past_key_values'] = past_key_values

        output = autoBertModel(inputs)

        prefixEncoderModel = Model(input_list,output.last_hidden_state)


        self.embed_model = prefixEncoderModel
        self.embedding_size = prefixEncoderModel.output.shape[-1]



if __name__ == '__main__':



    model_name = "roberta-large"
    a = AutoEmbeddingPrefixEmbedding(model_name,["input_ids","attention_mask"])
    a.build_embedding_model()

    config = AutoConfig.from_pretrained(model_name)
    model = TFAutoModel.from_pretrained(model_name)
    from transformers import BertTokenizer
    tokenizer = AutoTokenizer.from_pretrained(model_name)
    tf_batch = tokenizer(
        ['we are very happy to show you the transformers library.',
         "we hope you don't hate it.",
         ["he "] * 12
         ],
        padding=True,
        truncation=True,
        max_length=128,
        return_tensors='tf',
    )
    #tf_outputs = model(tf_batch)
    tf_outputs = model(tf_batch, attention_mask = tf_batch['attention_mask'],labels=tf.constant([1, 0]))
    print(tf_outputs)


from tensorflow.keras.utils import get_file
from typing import Tuple, List
from nlp_tools import macros as k
from nlp_tools.corpus import DataReader
from nlp_tools import utils

import logging
import os
import glob
from tqdm import tqdm
import json
from nlp_tools.utils.ner_utils import get_entities

class ChineseDailyNerCorpus(object):
    """
        Chinese Daily New New Corpus
        https://github.com/zjy-ucas/ChineseNER/
        """
    __corpus_name__ = 'china-people-daily-ner-corpus'
    __zip_file__name = 'http://s3.bmio.net/kashgari/china-people-daily-ner-corpus.tar.gz'
    @classmethod
    def load_data(cls,
                  subset_name:str = 'train',
                  shuffle:bool = True,
                  corpus_path = None) -> Tuple[List[List[str]],List[List[str]]]:
        """
                Load dataset as sequence labeling format, char level tokenized

                features: ``[['海', '钓', '比', '赛', '地', '点', '在', '厦', '门', ...], ...]``

                labels: ``[['O', 'O', 'O', 'O', 'O', 'O', 'O', 'B-LOC', 'I-LOC', ...], ...]``

                Sample::

                    train_x, train_y = ChineseDailyNerCorpus.load_data('train')
                    test_x, test_y = ChineseDailyNerCorpus.load_data('test')

                Args:
                    subset_name: {train, test, valid}
                    shuffle: should shuffle or not, default True.

                Returns:
                    dataset_features and dataset labels
                """
        if corpus_path == None:
            corpus_path = get_file(cls.__corpus_name__,
                                   cls.__zip_file__name,
                                   cache_dir=k.DATA_PATH,
                                   untar=True)
        if subset_name == 'train':
            file_path = os.path.join(corpus_path,'example.train')
        elif subset_name == 'test':
            file_path = os.path.join(corpus_path,'example.test')
        elif subset_name == 'valid':
            file_path = os.path.join(corpus_path,'example.dev')
        else:
            file_path = os.path.join(corpus_path,subset_name)

        x_data,y_data = DataReader.read_conll_format_file(file_path)
        if shuffle:
            x_data,y_data = utils.unison_shuffled_copies(x_data,y_data)
        logging.debug(f"loaded {len(x_data)} samples from {file_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")

        data = []
        for x,y in zip(x_data,y_data):
            item = []
            item.append("".join(x))
            y_change = get_entities(y)
            y_change = [(start,end,label) for (label,start,end) in y_change]
            item.append(y_change)

            data.append(item)


        return data


class CONLL2003ENCorpus(object):
    __corpus__name__ = 'conll2003_en'
    __zip_file__name = 'http://s3.bmio.net/kashgari/conll2003_en.tar.gz'

    @classmethod
    def load_data(cls,
                  subset_name: str = 'train',
                  task_name: str = 'ner',
                  shuffle: bool = True) -> Tuple[List[List[str]],List[List[str]]]:
        corpus_path = get_file(cls.__corpus__name__,
                               cls.__zip_file__name,
                               cache_dir=k.DATA_PATH,
                               untar=True)

        if subset_name not in {'train', 'test', 'valid'}:
            raise ValueError()

        file_path = os.path.join(corpus_path, f'{subset_name}.txt')

        if task_name not in {'pos', 'chunking', 'ner'}:
            raise ValueError()

        data_index = ['pos', 'chunking', 'ner'].index(task_name) + 1

        x_data, y_data = DataReader.read_conll_format_file(file_path, label_index=data_index)
        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)
        logging.debug(f"loaded {len(x_data)} samples from {file_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")
        return x_data, y_data


class NerTwoLineCorpus(object):
    '''
        通用数据格式加载：
        x x x x x x x \n
        O O B-XXX I-XXX O O O \n
        \n\n
    '''
    __corpus__name__ = '通用NER数据加载类'

    @classmethod
    def load_data(cls,
                  glob_corpus_path,
                  subset_name: str = 'train',
                  task_name: str = 'ner',
                  shuffle: bool = True) -> Tuple[List[List[str]], List[List[str]]]:


        if subset_name not in {'train', 'test', 'valid'}:
            raise ValueError()


        if task_name not in {'pos', 'chunking', 'ner'}:
            raise ValueError()

        file_list = glob.glob(glob_corpus_path)
        x_data, y_data = [] ,[]
        for file in tqdm(file_list):
            with open(file,encoding='utf-8') as fread:
                lines = fread.read()
                train_lists = lines.split("\n\n")

                for item in train_lists:
                    if not item.strip():
                        continue
                    words_and_tags = item.strip().split("\n")
                    if len(words_and_tags) != 2:
                        print(words_and_tags)
                    else:
                        x_data.append(words_and_tags[0])#.split(" "))
                        y_data.append(words_and_tags[1])#.split(" "))



        if shuffle:
            x_data, y_data = utils.unison_shuffled_copies(x_data, y_data)
        logging.debug(f"loaded {len(x_data)} samples from {glob_corpus_path}. Sample:\n"
                      f"x[0]: {x_data[0]}\n"
                      f"y[0]: {y_data[0]}")
        return x_data, y_data



class JsonNerCorpus(object):
    '''

    '''

    def __init__(self,end_position_plus=False):
        self.end_position_plus = end_position_plus

    def generate_examples(self, filepath):
        with open(filepath, 'r', encoding="utf-8") as f:
            for line in f:
                if line != "" and line != "\n":
                    line_dict = json.loads(line.strip())
                    text = line_dict['text']
                    labels = line_dict['labels']
                    if not self.end_position_plus:
                        tuple_lables = [tuple(item) for item in labels]
                    else:
                        tuple_lables = [tuple([item[0],item[1]-1,item[2]]) for item in labels]
                    yield text, tuple_lables

    def load_data(self,glob_path):
        data = []
        for file in glob.glob(glob_path):
            for x,y in self.generate_examples(file):
                data.append((x,y))
        return data

if __name__ == "__main__":
    glob_path = '/home/qiufengfeng/nlp/nlp_project/github_third/TransformersNer/ner_train/data/*test.json'
    data = JsonNerCorpus().load_data(glob_path)
    print(data)




from tensorflow.keras.layers import Layer,Dense,LayerNormalization,Concatenate

class PoolerInputLayer(Layer):
    def __init__(self, hidden_size,output_size):
        super(Layer, self).__init__()
        self.dense_0 = Dense(hidden_size,activation='tanh')
        self.LayerNorm = LayerNormalization()
        self.dense_1 = Dense( output_size)

    def __call__(self,inputs,mask=None, **kwargs):
        input = Concatenate()(inputs)
        x = self.dense_0(input)
        x = self.LayerNorm(x)
        x = self.dense_1(x)
        return x


import json
import  os
import pathlib
import time
from typing import Union,Any
import numpy as np

from nlp_tools.tasks.abs_task_model import ABCTaskModel
from bert4keras.backend import  search_layer
import tensorflow.keras.backend as K
import tensorflow as tf


def convert_to_saved_model(model: ABCTaskModel,
                           model_path: str,
                           version: Union[str, int] = None,
                           signatures: Any = None,
                           options: Any = None) -> None:
    """
    Export model for tensorflow serving
    Args:
        model: Target model.
        model_path: The path to which the SavedModel will be stored.
        version: The model version code, default timestamp
        signatures: Signatures to save with the SavedModel. Applicable to the
            'tf' format only. Please see the `signatures` argument in
            `tf.saved_model.save` for details.
        options: Optional `tf.saved_model.SaveOptions` object that specifies
            options for saving to SavedModel.

    """
    if not isinstance(model,ABCTaskModel):
        raise ValueError("Only supports the classification model and labeling model")
    if version is None:
        version = round(time.time())
    export_path = os.path.join(model_path,str(version))

    pathlib.Path(export_path).mkdir(exist_ok=True,parents=True)
    model.tf_model.save(export_path,save_format='tf',signatures=signatures,options=options)

    with open(os.path.join(export_path,'model_config.json'),'w') as f:
        f.write(json.dumps(model.to_dict(),indent=2,ensure_ascii=True))
        f.close()


def adversarial_training(model, embedding_name, epsilon=1):
    """给模型添加对抗训练
    其中model是需要添加对抗训练的keras模型，embedding_name
    则是model里边Embedding层的名字。要在模型compile之后使用。
    """

    if model.train_function is None:  # 如果还没有训练函数
        model.make_train_function()  # 手动make
    old_train_function = model.train_function  # 备份旧的训练函数

    # 查找Embedding层
    for output in model.outputs:
        embedding_layer = search_layer(output, embedding_name)
        if embedding_layer is not None:
            break
    if embedding_layer is None:
        raise Exception('Embedding layer not found')

    # 求Embedding梯度
    embeddings = embedding_layer.embeddings  # Embedding矩阵
    with tf.GradientTape() as gtape:
        gradients = gtape.gradient(model.compiled_loss._user_losses, [embeddings])  # Embedding梯度
        gradients = tf.zeros_like(embeddings) + gradients[0]  # 转为dense tensor

        # 封装为函数
        inputs = (
            model._feed_inputs + model._feed_targets + model._feed_sample_weights
        )  # 所有输入层
        embedding_gradients = tf.function(
            inputs=inputs,
            outputs=[gradients],
            name='embedding_gradients',
        )  # 封装为函数

    def train_function(inputs):  # 重新定义训练函数
        grads = embedding_gradients(inputs)[0]  # Embedding梯度
        delta = epsilon * grads / (np.sqrt((grads**2).sum()) + 1e-8)  # 计算扰动
        K.set_value(embeddings, K.eval(embeddings) + delta)  # 注入扰动
        outputs = old_train_function(inputs)  # 梯度下降
        K.set_value(embeddings, K.eval(embeddings) - delta)  # 删除扰动
        return outputs

    model.train_function = train_function  # 覆盖原训练函数


from abc import ABC
from typing import Dict,Any
import tensorflow as tf
import tensorflow.keras.layers  as L
from tensorflow import keras
from nlp_tools.logger import logger

from nlp_tools.tasks.classification.abc_model import ABCClassificationModel

class CNN_Attention_Model(ABCClassificationModel, ABC):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'conv_layer1': {
                'filters': 264,
                'kernel_size': 3,
                'padding': 'same',
                'activation': 'relu'
            },
            'conv_layer2': {
                'filters': 128,
                'kernel_size': 3,
                'padding': 'same',
                'activation': 'relu'
            },
            'conv_layer3': {
                'filters': 64,
                'kernel_size': 3,
                'padding': 'same',
                'activation': 'relu'
            },
            'layer_output': {
            },
        }

    def build_model_arc(self) -> None:
        if tuple(tf.__version__.split('.')) < tuple('2.1.0'.split('.')):
            logger.warning("Attention layer not serializable because it takes init args "
                           "but doesn't implement get_config. "
                           "Please try Attention layer with tf versions >= 2.1.0. "
                           "Issue: https://github.com/tensorflow/tensorflow/issues/32662")
        output_dim = self.label_processor.vocab_size
        config = self.hyper_parameters

        embed_model = self.embedding.embed_model
        # Query embeddings of shape [batch_size, Tq, dimension].
        query_embeddings = embed_model.output
        # Value embeddings of shape [batch_size, Tv, dimension].
        value_embeddings = embed_model.output

        # CNN layer.
        cnn_layer_1 = L.Conv1D(**config['conv_layer1'])
        # Query encoding of shape [batch_size, Tq, filters].
        query_seq_encoding = cnn_layer_1(query_embeddings)
        # Value encoding of shape [batch_size, Tv, filters].
        value_seq_encoding = cnn_layer_1(value_embeddings)

        cnn_layer_2 = L.Conv1D(**config['conv_layer2'])
        query_seq_encoding = cnn_layer_2(query_seq_encoding)
        value_seq_encoding = cnn_layer_2(value_seq_encoding)

        cnn_layer_3 = L.Conv1D(**config['conv_layer3'])
        query_seq_encoding = cnn_layer_3(query_seq_encoding)
        value_seq_encoding = cnn_layer_3(value_seq_encoding)

        # Query-value attention of shape [batch_size, Tq, filters]
        query_value_attention_seq = L.Attention()(
            [query_seq_encoding, value_seq_encoding])

        # Reduce over the sequence axis to produce encodings of shape
        # [batch_size, filters].
        query_encoding = L.GlobalMaxPool1D()(query_seq_encoding)
        query_value_attention = L.GlobalMaxPool1D()(query_value_attention_seq)

        # Concatenate query and document encodings to produce a DNN input layer.
        input_layer = L.Concatenate(axis=-1)([query_encoding, query_value_attention])

        output = L.Dense(output_dim, **config['layer_output'])(input_layer)
        output = self._activation_layer()(output)

        self.tf_model = keras.Model(embed_model.input, output)


if __name__ == "__main__":
    pass



from typing import Callable
from typing import Dict,Any

import numpy as np
from sklearn import metrics
from nlp_tools.tasks.base_model import BaseModel

class BaseScoringModel(BaseModel):
    """Base Sequence Labeling Model"""

    __task__ = 'scoring'

    @classmethod
    def get_default_hyper_parameters(cls) -> Dict[str,Dict[str,Any]]:
        raise NotImplementedError

    def compile_model(self,**kwargs):
        if kwargs.get("loss") is None:
            kwargs['loss'] = 'mse'

        if kwargs.get("optimizer") is None:
            kwargs['optimizer'] = 'rmsprop'

        if kwargs.get("metrics") is None:
            kwargs['metrics'] = ['mae']
        super(BaseScoringModel,self).compile_model(**kwargs)


    def evaluate(self,
                 x_data,
                 y_data,
                 batch_size = None,
                 should_round:bool =False,
                 round_func:Callable = None,
                 digits=4,
                 debug_info=False) -> Dict:
        """
                Build a text report showing the main classification metrics.

                Args:
                    x_data:
                    y_data:
                    batch_size:
                    should_round:
                    round_func:
                    digits:
                    debug_info:

                Returns:

                """
        y_pred = self.predict(x_data,batch_size=batch_size)

        if should_round:
            if round_func is None:
                round_func = np.round
            print(self.processor.output_dim)
            if self.processor.output_dim != 1:
                raise ValueError("Evaluate with round fuction only accept 1D output")
            y_pred = [round_func(i) for i in y_pred]
            report = metrics.classification_report(y_data,
                                                   y_pred,
                                                   digits=digits)
            report_dict = metrics.classification_report(y_data,
                                                        y_pred,
                                                        output_dict=True,
                                                        digits=digits)
            print(report)
        else:
            mean_squared_error = metrics.mean_squared_error(y_data,y_pred)
            r2_score = metrics.r2_score(y_data,y_pred)
            report_dic = {
                'mean_squared_error': mean_squared_error,
                'r2_score': r2_score
            }
            print(f"mean_squared_error : {mean_squared_error}\n"
                  f"r2_score           : {r2_score}")
        return report_dic


from abc import ABC
from typing import Dict,List,Any,Tuple
import numpy as np

from nlp_tools.generators import  CorpusGenerator
from nlp_tools.types import TextSamplesVar



class ABCLabelProcessor(ABC):
    def to_dict(self) -> Dict[str, Any]:
        return {
            'config': {
                'vocab2idx': self.vocab2idx,
                'max_sentence_length':self.max_sentence_length,
                "embedding_max_position":self.embedding_max_position,
                "token_pad":self.token_pad
            },
            '__class_name__': self.__class__.__name__,
            '__module__': self.__class__.__module__,
        }

    def __init__(self,max_sentence_length=None,embedding_max_position=None,vocab2idx={},**kwargs: Any) -> None:
        self.vocab2idx = vocab2idx
        self.idx2vocab = dict([(v, k) for k, v in self.vocab2idx.items()])
        self.max_sentence_length = max_sentence_length
        self.embedding_max_position = embedding_max_position
        self.token_pad = "[PAD]"


    @property
    def vocab_size(self) -> int:
        return len(self.vocab2idx)

    @property
    def is_vocab_build(self) -> bool:
        return self.vocab_size != 0

    def build_vocab(self,
                    x_data: TextSamplesVar,
                    y_data: TextSamplesVar) -> None:
        corpus_gen = CorpusGenerator(x_data, y_data)
        self.build_vocab_generator([corpus_gen])

    def build_vocab_generator(self,
                              generators: List[CorpusGenerator]) -> None:
        raise NotImplementedError

    def get_tensor_shape(self, batch_size: int, seq_length: int) -> Tuple:
        return batch_size, seq_length

    def transform(self,samples,seq_length = None):
        raise NotImplementedError

    def inverse_transform(self,
                          labels,
                          lengths = None,
                          mapping_list=None,
                          threshold= 0.5) :
        raise NotImplementedError

    def update_length_info(self,embedding_max_position=None,max_sentence_length=None):
        self.embedding_max_position = embedding_max_position
        self.max_sentence_length = max_sentence_length






from typing import Dict, Any

from tensorflow import keras

from nlp_tools.layers import L
from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel


class BiLSTM_Model(ABCLabelingModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'layer_blstm': {
                'units': 128,
                'return_sequences': True
            },
            'layer_dropout': {
                'rate': 0.4
            },
            'layer_time_distributed': {},
        }

    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model

        layer_stack = [
            L.Bidirectional(L.LSTM(**config['layer_blstm']), name='layer_blstm'),
            L.Dropout(**config['layer_dropout'], name='layer_dropout'),
            L.Dense(output_dim, **config['layer_time_distributed']),
        ]
        tensor = embed_model.output
        for layer in layer_stack:
            tensor = layer(tensor)

        self.tf_model = keras.Model(embed_model.inputs, tensor)



from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer
from typing import Dict,Any


from transformers import AutoTokenizer,BertTokenizer



class HuggingTokenizer(ABCTokenizer):
    """hugging tokenizer wrapper
    """

    def to_dict(self) -> Dict[str, Any]:
        data = super(HuggingTokenizer, self).to_dict()
        data['config']['model_path'] = self.model_path
        data['config']['need_attention_ids'] = self.need_attention_ids

        self.tokenizer.save_pretrained(self.model_path)
        return data

    def __init__(self, model_path, need_attention_ids=True):
        super().__init__()
        self.model_path = model_path
        self._tokenizer:BertTokenizer = AutoTokenizer.from_pretrained(model_path)

        self.need_attention_ids = need_attention_ids

    def change_model_path(self, new_model_path):
        self.model_path = new_model_path

    def tokenize(self, text, maxlen=None, add_special_tokens=False,**kwargs):
        text = self.tokenizer.tokenize(text,add_special_tokens=add_special_tokens,max_length=maxlen)
        return text


    def encode(self,
                text_or_list,
                second_text=None,
                maxlen=None,
                is_split_into_words=False,
                add_special_tokens=True,
               **kwargs):
        if type(text_or_list) == str:
            text_or_list = [text_or_list]


        tokener_dict = self.tokenizer(text_or_list,padding=True,truncation=True,return_tensors='np',max_length=maxlen,is_split_into_words=is_split_into_words,add_special_tokens=add_special_tokens,**kwargs)


        if not self.need_attention_ids:
            if 'attention_mask' in tokener_dict:
                del tokener_dict['attention_mask']
        return [value for key,value in tokener_dict.items()]
    





from bert4keras.snippets import AutoRegressiveDecoder
from tensorflow.keras.models import Model
import numpy as np

class AutoTitle(AutoRegressiveDecoder):
    """解码器
    """

    def __init__(self,generator_mode,tokenizer,max_generator_len,max_seq_len=32 ,**kwargs):
        kwargs["maxlen"] = max_generator_len
        super(AutoTitle,self).__init__(**kwargs)
        self.generate_mode = generator_mode
        self.tokenizer = tokenizer
        self.max_seq_len = max_seq_len
        self.max_generator_len = max_generator_len


    @AutoRegressiveDecoder.wraps(default_rtype='probas')
    def predict(self, inputs, output_ids, states):
        token_ids, segment_ids = inputs
        token_ids = np.concatenate([token_ids, output_ids], 1)
        segment_ids = np.concatenate([segment_ids, np.ones_like(output_ids)], 1)
        if type(self.generate_mode) != Model:
            return self.last_token(self.generate_mode.tf_model).predict([token_ids, segment_ids])[:, -1]
        else:
            return self.last_token(self.generate_mode).predict([token_ids, segment_ids])[:, -1]

    def generate(self, text, topk=1):
        token_ids, segment_ids = self.tokenizer.encode(text, maxlen=self.max_seq_len)
        output_ids = self.beam_search([token_ids, segment_ids],
                                      topk=topk)  # 基于beam search
        return self.tokenizer.decode(output_ids)



from nlp_tools.tokenizer.base_tokenizer import ABCTokenizer


class ListTokenizer(ABCTokenizer):
    """list tokenizer
    """

    def __init__(self):
        pass

    def tokenize(self, text: str, **kwargs):
        """
        Tokenize text into token sequence
        Args:
            text: target text sample

        Returns:
            List of tokens in this sample
        """

        return list(text)

    def encode(self, text: str, **kwargs):
        """
        Tokenize text into token sequence
        Args:
            text: target text sample

        Returns:
            List of tokens in this sample
        """

        return list(text)

if __name__ == '__main__':
    a = ListTokenizer()
    print(a.to_dict())


#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   __init__.py.py    
@Contact :   544855237@qq.com
@License :   (C)Copyright 2017-2018, Liugroup-NLPR-CASIA

@Modify Time      @Author    @Version    @Desciption
------------      -------    --------    -----------
2021/12/20 下午3:24   qiufengfeng      1.0         None
'''

# import lib
 



from tensorflow.keras.layers import Layer
from tensorflow.keras import backend as K
import nlp_tools
import tensorflow as tf

class NonMaskingLayer(Layer):
    def __init__(self,**kwargs):
        self.supports_masking = True
        super(NonMaskingLayer,self).__init__(**kwargs)

    def build(self, input_shape):
        pass

    def compute_mask(self, inputs, mask=None):
        return None
    def call(self,x,mask=None):
        return x


class MaskingSaverLayer(Layer):
    def __init__(self,**kwargs):
        self.supports_masking = True
        self.mask = None
        self.loss_func = None
        super(MaskingSaverLayer,self).__init__(**kwargs)

    def build(self, input_shape):
        pass

    def compute_mask(self, inputs, mask=None):
        return None
    def call(self,x,mask=None):
        self.mask = mask
        return x
    def set_loss_func(self,loss_func):
        self.loss_func = loss_func
    def loss(self, y_true, y_pred):
        loss_ = self.loss_func(y_true, y_pred)
        mask = tf.cast(self.mask, dtype=loss_.dtype)  # 将前面统计的是否零转换成1，0的矩阵
        loss_ *= mask  # 将正常计算的loss加上mask的权重，就剔除了padding 0的影响
        loss_ = tf.math.divide_no_nan(tf.reduce_sum(loss_, axis=-1), tf.reduce_sum(mask, axis=-1))
        return loss_



nlp_tools.custom_objects['NonMaskingLayer'] = NonMaskingLayer
nlp_tools.custom_objects['MaskingSaverLayer'] = MaskingSaverLayer


import warnings
import tensorflow as tf
from typing import TYPE_CHECKING,Union
from tensorflow.keras.utils import CustomObjectScope

from nlp_tools import  custom_objects
from nlp_tools.utils.serialize import load_data_object
from nlp_tools.utils.data import get_list_subset
from nlp_tools.utils.data import unison_shuffled_copies
from nlp_tools.utils.multi_label import MultiLabelBinarizer



def custom_object_scope() -> CustomObjectScope:
    return tf.keras.utils.custom_object_scope(custom_objects)




import os
from pathlib import Path
from typing import Dict

DATA_PATH = os.path.join(str(Path.home()), '.nlp_tools')

Path(DATA_PATH).mkdir(exist_ok=True, parents=True)



class Config(object):

    def __init__(self):
        self.verbose = False

    def to_dict(self):
        return {
            'verbose': self.verbose
        }


config = Config()

if __name__ == "__main__":
    print("Hello world")



from typing import Dict, Any

from tensorflow import keras
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.metrics import Accuracy

from nlp_tools.tasks.labeling.abc_model import ABCLabelingModel
from nlp_tools.layers.position_layer import GloablPointerLayer
from nlp_tools.loss.multilabel_loss import global_pointer_crossentropy
from nlp_tools.metrics.ner.global_pointer import global_pointer_f1_score

from nlp_tools.utils.ner_utils import output_ner_results



class GlobalPointModel(ABCLabelingModel):

    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {
            'global_point_layer': {
                'head_size': 64,
                'RoPE': True
            }
        }

    def build_model_arc(self) -> None:
        label_size = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model
        tensor = embed_model.output
        tensor = keras.layers.Dropout(0.5)(tensor)
        output = GloablPointerLayer(label_size,**config['global_point_layer'])(tensor)
        self.tf_model = keras.Model(embed_model.inputs, output)
        self.tf_model.summary()


    def compile_model(self,
                      loss: Any = None,
                      optimizer: Any = None,
                      metrics: Any = None,
                      **kwargs: Any) -> None:
        if loss is None:
            loss = global_pointer_crossentropy
        if metrics is None:
            metrics = global_pointer_f1_score

        if optimizer is None:
            optimizer = Adam(learning_rate=2e-5)

        self.tf_model.compile(loss=loss,
                              optimizer=optimizer,
                              metrics=metrics,
                              **kwargs)


    def predict(self,
                x_data,
                batch_size: int = 32,
                truncating: bool = False,
                predict_kwargs=None,
                return_max_entity=False) :
        if predict_kwargs is None:
            predict_kwargs = {}

        if truncating:
            seq_length = self.max_sequence_length
        else:
            seq_length = None

        if type(x_data[0]) == list:
            x_data = ["".join(x).replace("##","").replace('[CLS]',"").replace("[SEP]","") for x in x_data]

        x_data_tokenized = [self.text_processor.text_tokenizer.tokenize(x) for x in x_data]
        lengths = [len(x) for x in x_data_tokenized]

        tensor = self.text_processor.transform(x_data,seq_length=seq_length)
        pred = self.tf_model.predict(tensor, batch_size=batch_size, verbose=1, **predict_kwargs)

        x_data_mapping = [self.text_processor.text_tokenizer.rematch(x, x_tokens) for x, x_tokens in
                          zip(x_data, x_data_tokenized)]
        res = self.label_processor.inverse_transform(pred,lengths=lengths,mapping_list =x_data_mapping)
        return res


    def predict_entities(self,
                         x_data,
                         batch_size: int = 32,
                         truncating: bool = False,
                         predict_kwargs: Dict = None) :
        predict_result = self.predict(x_data,batch_size,truncating,predict_kwargs)

        result = output_ner_results(x_data,predict_result)
        return result

    def predict_max_entities(self,
                         x_data,
                         batch_size: int = 32,
                         truncating: bool = False,
                         predict_kwargs: Dict = None) :
        predict_result = self.predict(x_data,batch_size,truncating,predict_kwargs,return_max_entity=True)

        result = output_ner_results(x_data,predict_result)
        return result









from tensorflow import keras
import tensorflow as tf
import tensorflow.keras.backend as K
import numpy as np

from bert4keras.layers import Loss
from bert4keras.backend import search_layer



def contrastive_loss(y_true, y_pred):
    '''Contrastive loss from Hadsell-et-al.'06
        http://yann.lecun.com/exdb/publis/pdf/hadsell-chopra-lecun-06.pdf
        '''
    margin = 1
    sqaure_pred = tf.square(y_pred)
    margin_square = tf.square(tf.maximum(margin - y_pred, 0))
    return tf.reduce_mean(y_true * sqaure_pred + (1 - y_true) * margin_square)


class TextGenerateCrossEntropy(Loss):
    """交叉熵作为loss，并mask掉输入部分
     """
    def compute_loss(self, inputs, mask=None):
        y_true, y_mask, y_pred = inputs
        y_true = y_true[:, 1:]  # 目标token_ids
        y_mask = y_mask[:, 1:]  # segment_ids，刚好指示了要预测的部分
        y_pred = y_pred[:, :-1]  # 预测序列，错开一位
        loss = K.sparse_categorical_crossentropy(y_true, y_pred)
        loss = K.sum(loss * y_mask) / K.sum(y_mask)
        return loss


def categorical_focal_loss(alpha=0.25, gamma=2.):
    """
        Softmax version of focal loss.
        When there is a skew between different categories/labels in your data set, you can try to apply this function as a
        loss.
               m
          FL = ∑  -alpha * (1 - p_o,c)^gamma * y_o,c * log(p_o,c)
              c=1
          where m = number of classes, c = class and o = observation
        Parameters:
          alpha -- the same as weighing factor in balanced cross entropy. Alpha is used to specify the weight of different
          categories/labels, the size of the array needs to be consistent with the number of classes.
          gamma -- focusing parameter for modulating factor (1-p)
        Default value:
          gamma -- 2.0 as mentioned in the paper
          alpha -- 0.25 as mentioned in the paper
        References:
            Official paper: https://arxiv.org/pdf/1708.02002.pdf
            https://www.tensorflow.org/api_docs/python/tf/keras/backend/categorical_crossentropy
        Usage:
         model.compile(loss=[categorical_focal_loss(alpha=[[.25, .25, .25]], gamma=2)], metrics=["accuracy"], optimizer=adam)
        """
    alpha = np.array(alpha, dtype=np.float32)

    def categorical_focal_loss_fixed(y_true, y_pred):
        """
            :param y_true: A tensor of the same shape as `y_pred`
            :param y_pred: A tensor resulting from a softmax
            :return: Output tensor.
            """
        bh = K.prod(K.shape(y_pred)[:-1])
        y_true = K.reshape(y_true, (bh, -1))
        y_pred = K.reshape(y_pred, (bh, -1))


        # Clip the prediction value to prevent NaN's and Inf's
        epsilon = K.epsilon()
        y_pred = K.clip(y_pred,epsilon,1. - epsilon)

        # Calculate Cross Entropy
        y_true = K.cast(y_true,K.floatx())
        cross_entropy = -y_true * K.log(y_pred)

        # Calculate Focal Loss
        loss =  K.pow(1 - y_pred, gamma) * cross_entropy

        # Compute mean loss in mini_batch
        return K.mean(K.sum(loss,axis=-1))

    return categorical_focal_loss_fixed


class mutilplyLoss():
    __name__ = "mutilplyLoss"
    def __init__(self,loss_funs, epsilon=1):
        self.epsilon = epsilon
        self.loss_funs = loss_funs

    def __call__(self,y_true, y_pred):
        """带梯度惩罚的loss
        """
        loss = self.loss_funs(y_true, y_pred)
        embeddings = search_layer(y_pred, 'Embedding-Token').embeddings
        gp = K.sum(K.gradients(loss, [embeddings])[0].values ** 2)
        return loss + 0.5 * self.epsilon * gp


def loss_with_gradient_penalty(model,loss_func,epsilon=1):
    def loss_with_gradient_penalty_2(y_true, y_pred):
        loss = tf.math.reduce_mean(loss_func(y_true, y_pred))
        embeddings = search_layer(model, 'Embedding-Token').embeddings
        gp = tf.math.reduce_sum(tf.gradients(loss, [embeddings])[0].values**2)
        return loss + 0.5 * epsilon * gp
    return loss_with_gradient_penalty_2


def multi_category_focal_loss2(gamma=2., alpha=1):
    """
    focal loss for multi category of multi label problem
    适用于多分类或多标签问题的focal loss
    alpha控制真值y_true为1/0时的权重
        1的权重为alpha, 0的权重为1-alpha
    当你的模型欠拟合，学习存在困难时，可以尝试适用本函数作为loss
    当模型过于激进(无论何时总是倾向于预测出1),尝试将alpha调小
    当模型过于惰性(无论何时总是倾向于预测出0,或是某一个固定的常数,说明没有学到有效特征)
        尝试将alpha调大,鼓励模型进行预测出1。
    Usage:
     model.compile(loss=[multi_category_focal_loss2(alpha=0.25, gamma=2)], metrics=["accuracy"], optimizer=adam)
    """
    epsilon = 1.e-7
    gamma = float(gamma)
    alpha = tf.constant(alpha, dtype=tf.float32)

    def multi_category_focal_loss2_fixed(y_true, y_pred):
        y_true = tf.cast(y_true, tf.float32)
        y_pred = tf.clip_by_value(y_pred, epsilon, 1. - epsilon)

        alpha_t = y_true * alpha + (tf.ones_like(y_true) - y_true) * (1 - alpha)
        y_t = tf.multiply(y_true, y_pred) + tf.multiply(1 - y_true, 1 - y_pred)
        ce = -K.log(y_t)
        weight = tf.pow(tf.subtract(1., y_t), gamma)
        fl = tf.multiply(tf.multiply(weight, ce), alpha_t)
        loss = tf.reduce_mean(fl)
        return loss

    return multi_category_focal_loss2_fixed


def multi_category_focal_loss2_fixed(y_true, y_pred):
    epsilon = 1.e-7
    gamma=2.
    alpha = tf.constant(0.5, dtype=tf.float32)

    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.clip_by_value(y_pred, epsilon, 1. - epsilon)

    alpha_t = y_true*alpha + (tf.ones_like(y_true)-y_true)*(1-alpha)
    y_t = tf.multiply(y_true, y_pred) + tf.multiply(1-y_true, 1-y_pred)
    ce = -K.log(y_t)
    weight = tf.pow(tf.subtract(1., y_t), gamma)
    fl = tf.multiply(tf.multiply(weight, ce), alpha_t)
    loss = tf.reduce_mean(fl)
    return loss



import json
import os
import pathlib
from typing import Any, Tuple, List, Dict

import numpy as np
import tensorflow as tf
import tqdm

import nlp_tools
from nlp_tools.embeddings import BareEmbedding
from nlp_tools.embeddings.abc_embedding import  ABCEmbedding
from nlp_tools.generators import CorpusGenerator,Seq2SeqDataSet
from nlp_tools.logger import logger
from nlp_tools.processors import SequenceProcessor
from nlp_tools.tasks.seq2seq.decoder import AttGRUDecoder
from nlp_tools.tasks.seq2seq.encoder import GRUEncoder
from nlp_tools.types import TextSamplesVar

class Seq2Seq:
    def to_dict(self) -> Dict[str, Any]:
        return {
            'tf_version':tf.__version__,
            'nlp_tools_version':nlp_tools.__version__,
            '__class_name__':self.__class__.__name__,
            "__module__":self.__class__.__module__,
            'config':{
                'encoder_seq_length': self.encoder_seq_length,
                'decoder_seq_length': self.decoder_seq_length,
                'hidden_size':self.hidden_size
            },
            'encoder_embedding': self.encoder_embedding.to_dict(),  # type: ignore
            'decoder_embedding': self.decoder_embedding.to_dict(),
            'encoder_processor': self.encoder_processor.to_dict(),
            'decoder_processor': self.decoder_processor.to_dict(),
        }

    def __init__(self,
                 encoder_embedding: ABCEmbedding = None,
                 decoder_embedding: ABCEmbedding = None,
                 encoder_seq_length: int = None,
                 decoder_seq_length: int = None,
                 hidden_size: int = 1024,
                 **kwargs: Any):
        """
        Init Labeling Model

        Args:
            embedding: embedding object
            sequence_length: target sequence length
            hyper_parameters: hyper_parameters to overwrite
            **kwargs:
        """
        logger.warning("Seq2Seq API is experimental. It may be changed in the future without notice.")
        if encoder_embedding is None:
            encoder_embedding = BareEmbedding(embedding_size=256)  # type: ignore

        self.encoder_embedding = encoder_embedding

        if decoder_embedding is None:
            decoder_embedding = BareEmbedding(embedding_size=256)  # type: ignore

        self.decoder_embedding = decoder_embedding

        self.encoder_processor = SequenceProcessor(min_count=1)
        self.decoder_processor = SequenceProcessor(build_vocab_from_labels=True,min_count=1)

        self.encoder: GRUEncoder = None
        self.decoder: AttGRUDecoder = None

        self.hidden_size: int = hidden_size

        self.encoder_seq_length = encoder_seq_length
        self.decoder_seq_length = decoder_seq_length

        self.optimizer = tf.keras.optimizers.Adam()
        self.loss_object = tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True, reduction='none')

    # @tf.function
    def loss_function(self, real: tf.Tensor, pred: tf.Tensor) -> tf.Tensor:
        mask = tf.math.logical_not(tf.math.equal(real, 0))
        loss_ = self.loss_object(real, pred)

        mask = tf.cast(mask, dtype=loss_.dtype)
        loss_ *= mask

        return tf.reduce_mean(loss_)

    def build_model(self,
                    x_train: TextSamplesVar,
                    y_train: TextSamplesVar) -> None:
        train_gen = CorpusGenerator(x_train, y_train)
        self.build_model_generator(train_gen)

    def _build_encoder_decoder(self) -> None:
        self.encoder = GRUEncoder(self.encoder_embedding, hidden_size=self.hidden_size)
        self.decoder = AttGRUDecoder(self.decoder_embedding,
                                     hidden_size=self.hidden_size,
                                     vocab_size=self.decoder_processor.vocab_size)
        try:
            self.encoder.model().summary()
            self.decoder.model().summary()
        except:
            pass

    def build_model_generator(self,
                              train_gen: CorpusGenerator) -> None:
        """
        Build model with a generator, This function will do:

        1. setup processor's vocab if the vocab is empty.
        2. calculate the sequence length if `sequence_length` is None.
        3. build up model architect.
        4. compile the ``tf_model`` with default loss, optimizer and metrics.

        Args:
            train_gen: train data generator

        """
        if self.encoder is None:
            self.encoder_processor.build_vocab_generator([train_gen])
            self.decoder_processor.build_vocab_generator([train_gen])
            self.encoder_embedding.setup_text_processor(self.encoder_processor)
            self.decoder_embedding.setup_text_processor(self.decoder_processor)

            if self.encoder_seq_length is None:
                self.encoder_seq_length = self.encoder_embedding.get_seq_length_from_corpus([train_gen],
                                                                                            cover_rate=1.0)
                logger.info(f"calculated encoder sequence length: {self.encoder_seq_length}")

            if self.decoder_seq_length is None:
                self.decoder_seq_length = self.decoder_embedding.get_seq_length_from_corpus([train_gen],
                                                                                            use_label=True,
                                                                                            cover_rate=1.0)
                logger.info(f"calculated decoder sequence length: {self.decoder_seq_length}")

            self._build_encoder_decoder()

    # @tf.function
    def train_step(self,  # type: ignore
                   input_seq,
                   target_seq,
                   enc_hidden):
        loss = 0

        with tf.GradientTape() as tape:
            enc_output, enc_hidden = self.encoder(input_seq, enc_hidden)

            dec_hidden = enc_hidden

            bos_token_id = self.encoder_processor.vocab2idx[self.encoder_processor.token_bos]
            dec_input = tf.expand_dims([bos_token_id] * target_seq.shape[0], 1)

            # Teacher forcing - feeding the target as the next input
            for t in range(1, target_seq.shape[1]):
                # pass enc_output to the decoder
                predictions, dec_hidden, _ = self.decoder(dec_input, dec_hidden, enc_output)
                loss += self.loss_function(target_seq[:, t], predictions)
                # using teacher forcing
                dec_input = tf.expand_dims(target_seq[:, t], 1)

        batch_loss = (loss / int(target_seq.shape[1]))
        variables = self.encoder.trainable_variables + self.decoder.trainable_variables
        gradients = tape.gradient(loss, variables)
        self.optimizer.apply_gradients(zip(gradients, variables))

        return batch_loss

    def fit(self,
            x_train: TextSamplesVar,
            y_train: TextSamplesVar,
            *,
            batch_size: int = 64,
            epochs: int = 5,
            callbacks: List[tf.keras.callbacks.Callback] = None) -> tf.keras.callbacks.History:
        train_gen = CorpusGenerator(x_train, y_train)
        self.build_model_generator(train_gen)

        train_dataset = Seq2SeqDataSet(train_gen,
                                       batch_size=batch_size,
                                       encoder_processor=self.encoder_processor,
                                       encoder_seq_length=self.encoder_seq_length,
                                       decoder_processor=self.decoder_processor,
                                       decoder_seq_length=self.decoder_seq_length)

        if callbacks is None:
            callbacks = []
        history_callback = tf.keras.callbacks.History()
        callbacks.append(history_callback)

        for c in callbacks:
            c.set_model(self)
            c.on_train_begin()

        for epoch in range(epochs):
            for c in callbacks:
                c.on_epoch_begin(epoch=epoch)
            enc_hidden = tf.zeros((batch_size, self.hidden_size))
            total_loss = []

            with tqdm.tqdm(total=len(train_dataset)) as p_bar:
                for (inputs, targets) in train_dataset.take():
                    p_bar.update(1)
                    batch_loss = self.train_step(inputs, targets, enc_hidden)
                    total_loss.append(batch_loss.numpy())
                    info = f"Epoch {epoch + 1}/{epochs} | Epoch Loss: {np.mean(total_loss):.4f} " \
                           f"Batch Loss: {batch_loss.numpy():.4f}"
                    p_bar.set_description_str(info)
            logs = {'loss': np.mean(total_loss)}
            for c in callbacks:
                c.on_epoch_end(epoch=epoch, logs=logs)

        return history_callback

    def save(self, model_path: str) -> str:
        """
        Save model
        Args:
            model_path:
        """
        pathlib.Path(model_path).mkdir(exist_ok=True, parents=True)
        model_path = os.path.abspath(model_path)

        with open(os.path.join(model_path, 'model_config.json'), 'w') as f:
            f.write(json.dumps(self.to_dict(), indent=2, ensure_ascii=False))
            f.close()

        self.encoder_embedding.embed_model.save_weights(os.path.join(model_path, 'encoder_embed_weights.h5'))
        self.decoder_embedding.embed_model.save_weights(os.path.join(model_path, 'decoder_embed_weights.h5'))
        self.encoder.save_weights(os.path.join(model_path, 'encoder_weights.h5'))
        self.decoder.save_weights(os.path.join(model_path, 'decoder_weights.h5'))
        logger.info('model saved to {}'.format(os.path.abspath(model_path)))
        return model_path

    @classmethod
    def load_model(cls,model_path:str) -> 'Seq2Seq':
        from nlp_tools.utils import load_data_object
        model_config_path = os.path.join(model_path,'model_config.json')
        model_config = json.loads(open(model_config_path,'r',encoding='utf-8').read())
        model = load_data_object(model_config)

        # Load processors and embeddings
        model.encoder_processor = load_data_object(model_config['encoder_processor'])
        model.decoder_processor = load_data_object(model_config['decoder_processor'])
        model.encoder_embedding = load_data_object(model_config['encoder_embedding'])
        model.decoder_embedding = load_data_object(model_config['decoder_embedding'])

        model._build_encoder_decoder()
        # Load Model Weights
        model.encoder_embedding.embed_model.load_weights(os.path.join(model_path, 'encoder_embed_weights.h5'))
        model.decoder_embedding.embed_model.load_weights(os.path.join(model_path, 'decoder_embed_weights.h5'))

        # ------ Fix Start -------
        # load model issue on TF 2.3
        # Unable to load weights saved in HDF5 format into a subclassed Model which has not created its variables yet.
        # Call the Model first, then load the weights.
        input_seq = model.encoder_processor.transform([['hello']],
                                                      seq_length=model.encoder_seq_length)
        dec_input = tf.expand_dims([3], 0)
        enc_hidden = tf.zeros((1, model.hidden_size))
        dec_hidden = enc_hidden
        enc_output, enc_hidden = model.encoder(input_seq, enc_hidden)
        _ = model.decoder(dec_input, dec_hidden, enc_output)
        # ------ Fix End -------

        model.encoder.load_weights(os.path.join(model_path, 'encoder_weights.h5'))
        model.decoder.load_weights(os.path.join(model_path, 'decoder_weights.h5'))

        return model

    def predict(self,
                x_data: TextSamplesVar) -> Tuple[List, np.ndarray]:
        results = []
        attentions = []

        bos_token_id = self.decoder_processor.vocab2idx[self.decoder_processor.token_bos]
        eos_token_id = self.decoder_processor.vocab2idx[self.decoder_processor.token_eos]

        for sample in x_data:
            input_seq = self.encoder_processor.transform([sample], seq_length=self.encoder_seq_length)
            enc_hidden = tf.zeros((1, self.hidden_size))
            enc_output, enc_hidden = self.encoder(input_seq, enc_hidden)
            dec_hidden = enc_hidden

            attention_plot = np.zeros((self.decoder_seq_length, self.encoder_seq_length))
            token_out = []

            dec_input = tf.expand_dims([bos_token_id], 0)

            for t in range(self.decoder_seq_length):
                predictions, dec_hidden, att_weights = self.decoder(dec_input, dec_hidden, enc_output)
                # storing the attention weights to plot later on
                attention_weights = tf.reshape(att_weights, (-1,))
                attention_plot[t] = attention_weights.numpy()

                next_tokens = tf.argmax(predictions[0]).numpy()
                token_out.append(next_tokens)
                if next_tokens == eos_token_id:
                    break
                dec_input = tf.expand_dims([next_tokens], 0)
            r = self.decoder_processor.inverse_transform([token_out])[0]
            results.append(r)
            attentions.append(attention_plot)
        return results, np.array(attentions)


if __name__ == "__main__":
    from kashgari.corpus import ChineseDailyNerCorpus
    import logging

    logging.basicConfig(level='INFO', format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    x, y = ChineseDailyNerCorpus.load_data('test')
    x, y = x[:100], y[:100]

    seq2seq = Seq2Seq(hidden_size=256)
    seq2seq.build_model(x, y)
    seq2seq.save('./seq2seq_model')

    s = Seq2Seq.load_model('./seq2seq_model')
    res, att = seq2seq.predict(x[:10])
    res2, att2 = s.predict(x[:10])
    print(res == res2)
    print((att == att2).all())



from tensorflow import keras




class GlobalPointEvaluator(keras.callbacks.Callback):
    def __init__(self,model_check,model_save_path,valid_data,batch_size=32):
        self.model_check = model_check
        self.model_save_path = model_save_path
        self.best_val_f1 = 0
        self.batch_size= batch_size
        self.valid_x = [data[0] for data in valid_data]
        self.valid_y = [data[1] for data in valid_data]


    def caculate_f1(self):
        predict_labels  = self.model_check.predict(self.valid_x)
        X, Y, Z = 1e-10, 1e-10, 1e-10
        for predict_labels,real_labels in zip(predict_labels,self.valid_y):
            predict_labels = set(predict_labels)
            real_labels = set(real_labels)

            X += len(predict_labels & real_labels)
            Y += len(predict_labels)
            Z += len(real_labels)
        f1,precision,recall = round(2 * X/ (Y +Z),5),round(X/Y,5),round(X/Z,5)
        return f1,precision,recall


    def on_epoch_end(self, epoch, logs=None):
        f1,precision,recall = self.caculate_f1()

        if f1 >= self.best_val_f1:
            self.best_val_f1 = f1
            self.model_check.save(self.model_save_path)
        print(
            'valid:  f1: %.5f, precision: %.5f, recall: %.5f, best f1: %.5f\n' %
            (f1, precision, recall, self.best_val_f1)
        )


class GlobalPointCategoryEvaluator(keras.callbacks.Callback):
    def __init__(self,model_check,model_save_path,valid_data,batch_size=32):
        self.model_check = model_check
        self.model_save_path = model_save_path
        self.best_val_f1 = 0
        self.batch_size= batch_size
        self.valid_x = [data[0] for data in valid_data]
        self.valid_y = [data[1] for data in valid_data]


    def caculate_f1(self):
        predict_labels  = self.model_check.predict(self.valid_x)
        X, Y, Z = 1e-10, 1e-10, 1e-10
        for predict_labels,real_labels in zip(predict_labels,self.valid_y):
            predict_labels = set(predict_labels)
            real_labels = set(real_labels)

            X += len(predict_labels & real_labels)
            Y += len(predict_labels)
            Z += len(real_labels)
        f1,precision,recall = round(2 * X/ (Y +Z),5),round(X/Y,5),round(X/Z,5)
        return f1,precision,recall


    def on_epoch_end(self, epoch, logs=None):
        f1,precision,recall = self.caculate_f1()

        if f1 >= self.best_val_f1:
            self.best_val_f1 = f1
            self.model_check.save(self.model_save_path)
        print(
            'valid:  f1: %.5f, precision: %.5f, recall: %.5f, best f1: %.5f\n' %
            (f1, precision, recall, self.best_val_f1)
        )



import random
from typing import List,Union,TypeVar,Tuple

import numpy as np

T = TypeVar("T")


def get_list_subset(target: List[T], index_list: List[int]) -> List[T]:
    """
    Get the subset of the target list
    Args:
        target: target list
        index_list: subset items index

    Returns:
        subset of the original list
    """
    return [target[i] for i in index_list if i < len(target)]


def unison_shuffled_copies(a: List[T],
                           b: List[T]) -> Union[Tuple[List[T], ...], Tuple[np.ndarray, ...]]:
    """
    Union shuffle two arrays
    Args:
        a:
        b:

    Returns:

    """
    data_type = type(a)
    assert len(a) == len(b)
    c = list(zip(a, b))
    random.shuffle(c)
    a, b = zip(*c)
    if data_type == np.ndarray:
        return np.array(a), np.array(b)
    return list(a), list(b)


def load_vocab(dict_path, encoding='utf-8', startswith=None):
    """词典文件中读取词典
    """
    token_list = []
    with open(dict_path, encoding=encoding) as reader:
        for line in reader:
            token = line.strip()
            if token:
                token_list.append(token)
    if startswith:
        token_list = startswith + list(set(token_list))
    token_dict = {key:index for index,key in enumerate(token_list)}

    return token_dict


#!/usr/bin/env python
# -*- coding: UTF-8 -*-
'''
# author:fanfanfeng
# contact: 544855237@qq.com
# datetime:2022/2/14 下午11:31
'''

def get_list_start_position(main_list,sub_list,start = 0):
    '''
    找出sub_list在main_list里面的起始位置
    :param main_list:
    :param sub_list:
    :param start:main_list开始搜索的起始位置
    :return: 包含的下标（start_position,end_position）
    '''

    positions = []
    len1 = len(main_list)
    len2 = len(sub_list)
    if main_list == [] or sub_list == []:
        return -1
    for index in range(start,len1 - len2 + 1):
        if main_list[index] == sub_list[0] and main_list[index:index + len2] == sub_list:
            positions.append(index)
            positions.append(index + len2 -1)
            break


    return positions


def check_list_same(listA,listB):
    '''
    判断两个list是否相同
    :param listA:
    :param listB:
    :return: 1相同，0不相同
    '''
    the_same = True
    for itemA,itemB in zip(listA,listB):
        if itemA != itemB:
            the_same = False
            break
    return the_same



from typing import Dict,Any

from tensorflow.keras.layers import Flatten
from tensorflow.keras import layers
from tensorflow import keras
from tensorflow.keras import backend as K

from nlp_tools.layers import L
from nlp_tools.tasks.classification.abc_model import ABCClassificationModel
import tensorflow as tf
class ClassificationEntityLevel(ABCClassificationModel):
    @classmethod
    def default_hyper_parameters(cls) -> Dict[str, Dict[str, Any]]:
        return {

            'layer_output': {

            },
        }


    def build_model_arc(self) -> None:
        output_dim = self.label_processor.vocab_size

        config = self.hyper_parameters
        embed_model = self.embedding.embed_model


        entity_positon_inputs = L.Input(shape=(None,), dtype='int32')



        tensor = embed_model.output
        tensor = tf.gather(tensor, entity_positon_inputs, batch_dims=K.ndim(entity_positon_inputs) - 1)
        tensor = layers.Dropout(rate=0.3)(tensor)

        output_net = 'lstm'
        if output_net == "lstm":
            tensor = layers.LSTM(units=128)(tensor)#layers.Bidirectional(layers.LSTM(units=128))(tensor)
        else:
            #tensor = layers.GlobalAveragePooling1D()(tensor)
            tensor = layers.GlobalMaxPool1D()(tensor)


        #import tensorflow.keras.backend as K
        tensor = layers.Dropout(rate=0.3)(tensor)
        tensor = layers.Dense(128, activation=tf.nn.relu ,name="FFN/Dense")(tensor)
        tensor = L.Dense(output_dim, **config['layer_output'],name="FFN/Dense_classify")(tensor)
        output = self._activation_layer()(tensor)
        self.tf_model = keras.Model(embed_model.inputs + [entity_positon_inputs], output)





from tensorflow.keras.callbacks import Callback
import numpy as np
from sklearn.metrics import f1_score,classification_report,confusion_matrix

class F1CategoryCallback(Callback):
    def __init__(self,model_check,model_save_path,valid_data,batch_size=32):
        self.model_check = model_check
        self.model_save_path = model_save_path
        self.best_val_f1 = 0
        self.batch_size= batch_size
        self.valid_x = [x[0] for x in valid_data]
        self.valid_y = [x[1] for x in valid_data]
        self.count = 0


    def caculate_f1(self):
        predict_labels  = self.model_check.predict(self.valid_x,batch_size=self.batch_size)
        f1_score_value = f1_score(self.valid_y,predict_labels,average='macro')
        print(classification_report(self.valid_y,predict_labels))
        print(confusion_matrix(self.valid_y,predict_labels,labels=['不是',"不确定","是"]))

        return f1_score_value


    def on_epoch_end(self, epoch, logs=None):
        f1 = self.caculate_f1()

        if f1 >= self.best_val_f1:
            self.best_val_f1 = f1
            self.model_check.save(self.model_save_path)
            self.count = 0
        self.count += 1
        if self.count >15:
            self.model.stop_training = True
            self.count = 0
        print(
            'valid:  f1: %.5f, best f1: %.5f\n' %
            (f1, self.best_val_f1)
        )


import random
import json
class TianCiAddress2021():
    ## ccks2021 NLP https://tianchi.aliyun.com/competition/entrance/531901
    __corpus_name__ = 'ccks2021_address'

    @classmethod
    def load_data(cls,
                  file_path: str,
                  shuffle: bool = True,
                  data_agument=False):
        """
                {"text_id": "e225b9fd36b8914f42c188fc92e8918f", "query": "河南省巩义市新华路街道办事处桐和街6号钢苑新区3号楼一单元", "candidate": [{"text": "巩义市桐和街", "label": "不匹配"}, {"text": "桐和街依家小店", "label": "不匹配"}, {"text": "桐和街CHANG六LIULIU", "label": "不匹配"}
                """
        data = []
        with open(file_path, 'r', encoding='utf-8') as fread:
            for index, line in enumerate(fread):
                item_json = json.loads(line.strip())

                x1, x2, label = line.strip().split("\t")
                data.append(([x1, x2], str(label)))

        if shuffle:
            random.shuffle(data)
        return data


    def data_agument(self,json):
        pass



import pandas as pd
from nlp_tools import utils
import re


class DialogueCorpusLoader():
    __name__ ="dialogue_corpus_loader"
    def __init__(self,**kwargs):
        pass

    @classmethod
    def load_data(cls,data_path, shuffle: bool = True):
        df = pd.read_csv(data_path)
        sentences = df["data"].tolist()
        labels = df['label'].tolist()

        sentences_cut = []

        for sen in sentences:
            #sen = sen.replace("$","")
            sen_cut_list = sen.split("\n")
            sen_cut_list = [item for item in sen_cut_list if item != ""]

            sen_cut_list = [ re.sub("^.{0,1}\:","",sub_sen) for sub_sen in sen_cut_list]
            sentences_cut.append(sen_cut_list)

        if shuffle:
            sentences_cut, labels = utils.unison_shuffled_copies(sentences_cut, labels)

        data_union = [(x, y) for x, y in zip(sentences_cut, labels)]
        return data_union


if __name__ == '__main__':
    DialogueCorpusLoader.load_data("/home/fanfanfeng/working_data/nlp_data/working/classify/train.csv")










import codecs
import json
from typing import Dict,List,Any,Optional
from bert4keras.models import build_transformer_model
from nlp_tools.embeddings.abc_embedding import ABCEmbedding

class TransformerEmbedding(ABCEmbedding):
    """
    TransformerEmbedding is based on bert4keras.
    The embeddings itself are wrapped into our simple embedding interface so that they can be used like any other embedding.
    """
    def to_dict(self) -> Dict[str, Any]:
        info_dic = super(TransformerEmbedding, self).to_dict()
        info_dic['config']['model_type'] = self.model_type
        info_dic['config']['max_position'] = self.max_position
        info_dic['config']['config_path'] = self.config_path
        info_dic['config']['checkpoint_path'] = self.checkpoint_path
        info_dic['config']['bert_application'] = self.bert_application
        info_dic['config']['keep_tokens'] = self.bert_application
        return info_dic

    def __init__(self,
                 config_path=None,
                 checkpoint_path: str =None,
                 model_type: str = 'bert',
                 bert_application='encoder',
                 keep_tokens = None,
                 max_position=512,
                 **kwargs: Any):
        """

        Args:
            vocab_path: vocab file path, example `vocab.txt`
            config_path: model config path, example `config.json`
            checkpoint_path: model weight path, example `model.ckpt-100000`
            model_type: transfer model type, {bert, albert, nezha, gpt2_ml, t5}
            kwargs: additional params
        """
        super(TransformerEmbedding, self).__init__(**kwargs)

        self.checkpoint_path = checkpoint_path
        self.config_path = config_path

        self.model_type = model_type
        self.keep_tokens = keep_tokens
        self.bert_application = bert_application

        self.max_position = max_position


        self.with_pool = kwargs.get("with_pool",False)
        self.num_hidden_layers = kwargs.get("num_hidden_layers",12)
        # with open(config_path, 'r') as f:
        #     config = json.loads(f.read())
        #     if 'max_position' in config:
        #         self.max_position = config['max_position']
        #     else:
        #         self.max_position = config.get('max_position_embeddings')

        #super(TransformerEmbedding, self).__init__(**kwargs)



    def build_embedding_model(self) -> None:
        import tensorflow.keras.backend as K
        if self.model_type == 'electra':
            bert_model = build_transformer_model(config_path=self.config_path,
                                                 checkpoint_path=self.checkpoint_path,
                                                 model=self.model_type,
                                                 application=self.bert_application,
                                                 keep_tokens=self.keep_tokens,  # 只保留keep_tokens中的字，精简原字表
                                                 prefix="bert_encoder/"
                                                 )
        else:
            bert_model = build_transformer_model(config_path=self.config_path,
                                                 checkpoint_path=self.checkpoint_path,
                                                 model=self.model_type,
                                                 application=self.bert_application,
                                                 keep_tokens=self.keep_tokens,  # 只保留keep_tokens中的字，精简原字表
                                                 with_pool = self.with_pool,
                                                 num_hidden_layers = self.num_hidden_layers,
                                                 prefix="bert_encoder/"
                                                 )
        self.embed_model = bert_model
        self.embedding_size = bert_model.output.shape[-1]



